#include <vector>
#include <stdexcept>

#include "WolframLibrary.h"

DLLEXPORT mint WolframLibrary_getVersion(){
  return WolframLibraryVersion;
}

DLLEXPORT int WolframLibrary_initialize(WolframLibraryData){
  return LIBRARY_NO_ERROR;
}

DLLEXPORT void WolframLibrary_uninitialize(WolframLibraryData){
  return;
}

#include "QCD/scheme.hpp"
#include "QCD/threshold.hpp"
#include "QCD/alpha_s.hpp"
#include "QCD/MSbar_mass.hpp"

extern "C"{
  DLLEXPORT int
  alpha_s(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
  );

  DLLEXPORT int
  alpha_s_dec_up(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
  );

  DLLEXPORT int
  alpha_s_dec_down(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
  );

  DLLEXPORT int
  m(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
  );

  DLLEXPORT int
  m_dec_up(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
  );

  DLLEXPORT int
  m_dec_down(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
  );
}

namespace{
  QCD::strong_coupling get_alpha_s(
      WolframLibraryData,
      MArgument *Args
  ){
    const double alpha_0 = MArgument_getReal(Args[0]);
    const double mu_0 =  MArgument_getReal(Args[1]);
    const int nf =  MArgument_getInteger(Args[2]);
    const int running_order = MArgument_getInteger(Args[3]);
    return QCD::strong_coupling({alpha_0, mu_0, nf}, running_order);
  }

  QCD::MSbar_mass get_MSbar_mass(
      WolframLibraryData,
      MArgument *Args
  ){
    const double m_0 = MArgument_getReal(Args[0]);
    const double alpha_0 = MArgument_getReal(Args[1]);
    const double mu_0 =  MArgument_getReal(Args[2]);
    const int nf =  MArgument_getInteger(Args[3]);
    const int running_order = MArgument_getInteger(Args[4]);
    return QCD::MSbar_mass({m_0, mu_0, nf}, alpha_0, running_order);
  }

  std::vector<QCD::threshold> get_thresholds(
      WolframLibraryData libData,
      MArgument *Args
  ){
    const auto Mthresholds = MArgument_getMTensor(Args[0]);
    const auto Mdec_types = MArgument_getMTensor(Args[1]);
    const int Mthresholds_size =
      libData->MTensor_getFlattenedLength(Mthresholds)
      ;
    if(libData->MTensor_getType(Mthresholds) != MType_Real){
      throw std::invalid_argument(
	  "encountered decoupling threshold scale or mass with wrong type"
      );
    }
    if(libData->MTensor_getType(Mdec_types) != MType_Integer){
      throw std::invalid_argument(
	  "encountered decoupling with wrong type"
      );
    }
    if(Mthresholds_size % 2){
      throw std::invalid_argument(
	  "wrong format of decoupling threshold list: has odd length"
      );
    }
    if(Mthresholds_size/2 != libData->MTensor_getFlattenedLength(Mdec_types)){
      throw std::logic_error(
	  "size of threshold list and decoupling types do not match"
      );
    }
    std::vector<QCD::threshold> thresholds(Mthresholds_size/2);
    mint idx[1] = {0};
    for(int i = 1; i <= Mthresholds_size/2; ++i){
      mint dec_type;
      idx[0] = 2*i-1;
      assert(idx[0] < Mthresholds_size);
      libData->MTensor_getReal( Mthresholds, idx, &(thresholds[i-1].mu));
      idx[0] = 2*i;
      assert(idx[0] < Mthresholds_size);
      libData->MTensor_getReal( Mthresholds, idx, &(thresholds[i-1].m));
      idx[0] = i;
      libData->MTensor_getInteger( Mdec_types, idx, &dec_type);
      thresholds[i-1].dec_scheme = static_cast<QCD::scheme>(dec_type);
    };
    return thresholds;
  }
}

DLLEXPORT int
alpha_s(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 5){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    const double mu = MArgument_getReal(Args[0]);
    const auto alpha_s = get_alpha_s(libData, Args + 1);
    MArgument_setReal(Res, alpha_s(mu));
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
alpha_s_dec_up(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 7){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    const double mu = MArgument_getReal(Args[0]);
    auto alpha_s = get_alpha_s(libData, Args + 1);
    auto const thresholds = get_thresholds(libData, Args + 5);
    for(QCD::threshold const & thr: thresholds){
      alpha_s = alpha_s.decouple_up(thr);
    }
    MArgument_setReal(Res, alpha_s(mu));
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
alpha_s_dec_down(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 7){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    const double mu = MArgument_getReal(Args[0]);
    auto alpha_s = get_alpha_s(libData, Args + 1);
    auto const thresholds = get_thresholds(libData, Args + 5);
    for(QCD::threshold const & thr: thresholds){
      alpha_s = alpha_s.decouple_down(thr);
    }
    MArgument_setReal(Res, alpha_s(mu));
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
m(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 6){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    const double mu = MArgument_getReal(Args[0]);
    const auto m = get_MSbar_mass(libData, Args + 1);
    MArgument_setReal(Res, m(mu));
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
m_dec_up(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 8){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    const double mu = MArgument_getReal(Args[0]);
    auto m = get_MSbar_mass(libData, Args + 1);
    auto const thresholds = get_thresholds(libData, Args + 6);
    for(QCD::threshold const & thr: thresholds){
      m = m.decouple_up(thr);
    }
    MArgument_setReal(Res, m(mu));
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}

DLLEXPORT int
m_dec_down(
    WolframLibraryData libData,
    mint Argc, MArgument *Args,
    MArgument Res
){
  if(Argc != 8){
    return LIBRARY_FUNCTION_ERROR;
  }
  try{
    const double mu = MArgument_getReal(Args[0]);
    auto m = get_MSbar_mass(libData, Args + 1);
    auto const thresholds = get_thresholds(libData, Args + 6);
    for(QCD::threshold const & thr: thresholds){
      m = m.decouple_down(thr);
    }
    MArgument_setReal(Res, m(mu));
  }
  catch(std::exception & fail) {
    libData->Message(fail.what());
    return LIBRARY_FUNCTION_ERROR;
  }
  return LIBRARY_NO_ERROR;
}
