/**
 * @file
 * @brief ---
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "MSbar_mass.hpp"

#include <array>
#include <cmath>
#include <stdexcept>
#include <string>

#include "gamma_m.hpp"
#include "beta.hpp"
#include "array_2d.hpp"
#include "zeta_m.hpp"

namespace QCD{

  namespace{
    using detail::array_2d;

    constexpr unsigned nf_max = 6;

    //formulas according to arXiv:1402.6611 (Baikov, Chetyrkin, Kuehn)

    /*
       calculate d1,...,d3 according to eq. (4.12)
       note that the index of d is off by one!
    */
    array_2d<double, nf_max + 1, MSbar_mass::max_running_order> gen_d(){
      static constexpr size_t order = MSbar_mass::max_running_order;
      auto d = array_2d<double, nf_max + 1, order>();
      auto b = std::array<double, order + 1>();
      auto g = std::array<double, order + 1>();
      static_assert(b.size() > 3, "enough space for all d");
      double b0;

      for(size_t nf = 0; nf <= nf_max; ++nf){
	b0 = beta(0, nf);
	g[0] = gamma_m(0, nf)/b0;
	for(size_t i = 1; i < b.size(); ++i){
	  b[i] = beta(i, nf)/b0;
	  g[i] = gamma_m(i, nf)/b0;
	}
	d(nf, 0) = - b[1]*g[0] + g[1];
	d(nf, 1) = (- b[2]*g[0] + b[1]*(b[1]*g[0]-g[1]) + g[2])/2;
	d(nf, 2) = (
	    - b[3]*g[0] - b[2]*g[1]
	    + b[1]*(2*b[2]*g[0] - g[2] + b[1]*(g[1]-b[1]*g[0]))
	    + g[3]
	)/3;
      }
      return d;
    }

    // double d(int i, int nf){
    //   static const auto data = gen_d();
    //   assert(static_cast<unsigned>(nf) <= nf_max);
    //   assert(static_cast<unsigned>(i) <= MSbar_mass::max_running_order);
    //   return data(nf, i-1);
    // }

    //coefficients of c as in (4.8)
    auto gen_c() -> decltype(gen_d()){
      static const auto d = gen_d();
      static_assert(d.columns() >= 3, "enough coefficients d available");
      decltype(gen_d()) c;
      for(size_t nf = 0; nf <= nf_max; ++nf){
	c(nf, 0) = d(nf, 0);
	c(nf, 1) = d(nf, 1) + d(nf, 0)*d(nf, 0)/2;
	c(nf, 2) = d(nf, 2) + d(nf, 0)*(d(nf, 1) + d(nf, 0)*d(nf,0)/6);
      }
      return c;
    }

    double c_coeff(int i, int nf){
      static const auto data = gen_c();
      assert(static_cast<unsigned>(nf) <= nf_max);
      assert(static_cast<unsigned>(i) <= MSbar_mass::max_running_order);
      return data(nf, i);
    }

    //c as in (4.8)
    double c(double x, int nf, int order){
      if(static_cast<unsigned>(nf) > nf_max){
	throw std::out_of_range("number of flavours out of range");
      }
      double result = 0.;
      for(int o = order; o > 0; --o){
	result += c_coeff(o-1, nf);
	result *= x;
      }
      return pow(x, gamma_m(0, nf)/beta(0, nf)) * (1 + result);
    }
  }

  MSbar_mass::MSbar_mass(
      reference_mass ref,
      strong_coupling alpha_s,
      int run_order
  ): m0(ref.m), as0(alpha_s(ref.mu)), mu0(ref.mu),
     running_order(run_order),
     alpha_s_nf0(alpha_s) {
    if(order() < 0 || order() > max_running_order){
      throw std::out_of_range(
          "order " + std::to_string(order()) + " not in range 0..."
          + std::to_string(max_running_order)
      );
    }
    if(alpha_s_nf0.nf() != ref.nf){
      throw std::invalid_argument(
          "number of flavours " + std::to_string(alpha_s_nf0.nf())
          + " in strong coupling does not match number of flavours "
          + std::to_string(nf()) + " in mass"
      );
    }
  }

  MSbar_mass::MSbar_mass(
      reference_mass ref,
      double alpha_s_mu,
      int run_order
  ): m0(ref.m), as0(alpha_s_mu), mu0(ref.mu),
     running_order(run_order),
     alpha_s_nf0({alpha_s_mu, ref.mu, ref.nf}, run_order) {
    if(order() < 0 || order() > max_running_order){
      throw std::out_of_range(
          "order " + std::to_string(order()) + " not in range 0..."
          + std::to_string(max_running_order)
      );
    }
  }

  double MSbar_mass::operator()(double mu) const{
    assert(as0 == alpha_s_nf0(mu0));
    return evolve(m0, as0, alpha_s_nf0(mu), nf());
  }

  int MSbar_mass::order() const{
    return running_order;
  }

  int MSbar_mass::nf() const{
    return alpha_s_nf0.nf();
  }

  double MSbar_mass::to_scale_invariant(double ep) const{
    double m = m0;
    double mu = mu0;
    while(std::abs(m-mu) > ep){
      mu = m;
      m = operator()(mu);
    }
    return m;
  }

  // TODO: fix (sometimes? loops for ever)
  // double MSbar_mass::to_scale_invariant() const{
  //   double m = m0;
  //   double mu = mu0;
  //   while(
  //       (m > mu && m > std::nextafter(mu,m))
  //       || (m < mu && m < std::nextafter(mu,m))
  //   ){
  //     mu = m;
  //     m = operator()(mu);
  //   }
  //   return m;
  // }

  double MSbar_mass::evolve(
      double m_cur, double alpha_s_cur, double alpha_s_mu, int nf
  ) const{
    using detail::beta::pi;
    return c(alpha_s_mu/pi, nf, order())/c(alpha_s_cur/pi, nf, order())*m_cur;
  }

  double
  MSbar_mass::decouple_up(double alpha_s, threshold const & thr, int nf) const{
    switch(thr.dec_scheme){
    case scheme::pole:
      return zeta_m_inv<scheme::pole>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::MSbar:
      return zeta_m_inv<scheme::MSbar>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::scale_invariant:
      return zeta_m_inv<scheme::scale_invariant>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    default:
      throw std::invalid_argument("decoupling not implemented for scheme");
    }
  }

  double
  MSbar_mass::decouple_down(double alpha_s, threshold const & thr, int nf) const{
    switch(thr.dec_scheme){
    case scheme::pole:
      return zeta_m<scheme::pole>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::MSbar:
      return zeta_m<scheme::MSbar>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::scale_invariant:
      return zeta_m<scheme::scale_invariant>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    default:
      throw std::invalid_argument("decoupling not implemented for scheme");
    }
  }

  MSbar_mass
  MSbar_mass::remove(threshold thr) const{
    return decouple_down(thr);
  }

  MSbar_mass
  MSbar_mass::decouple_down(threshold thr) const{
    MSbar_mass const & m_nf = *this;
    MSbar_mass m_nfm1 = m_nf;
    m_nfm1.mu0 = thr.mu;
    m_nfm1.m0 = m_nf(thr.mu)*decouple_down(alpha_s_nf0(thr.mu), thr, nf());
    m_nfm1.alpha_s_nf0 = alpha_s_nf0.decouple_down(thr);
    m_nfm1.as0 = m_nfm1.alpha_s_nf0(thr.mu);
    return m_nfm1;
  }

  MSbar_mass
  MSbar_mass::add(threshold thr) const{
    return decouple_up(thr);
  }

  MSbar_mass MSbar_mass::decouple_up(threshold thr) const{
    MSbar_mass const & m_nf = *this;
    MSbar_mass m_nfp1 = m_nf;
    m_nfp1.mu0 = thr.mu;
    m_nfp1.m0 = m_nf(thr.mu)*decouple_up(alpha_s_nf0(thr.mu), thr, nf());
    m_nfp1.alpha_s_nf0 = alpha_s_nf0.decouple_up(thr);
    m_nfp1.as0 = m_nfp1.alpha_s_nf0(thr.mu);
    return m_nfp1;
  }
}
