/**
 * @file
 * @brief Corrections to the Green function
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "parameters.hpp"
#include "pole_resummation.hpp"

namespace QQbar_threshold{
  namespace detail{
    typedef std::complex<double> complex;

    //! Higher-order PNRQCD corrections to the Green function
    /**
     * @tparam N      Perturbative order. Leading order is N = 0.
     * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p       Further parameters.
     * @returns       The Nth order correction to the Green function.
     */
    template<int N>
    complex Green_function(complex lam, parameters const & p);

    //! Green function corrections with pole resummation
    /**
     * @tparam N      Perturbative order. Leading order is N = 0.
     * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p       Further parameters.
     * @returns       The Nth order correction to the Green function including pole resummation.
     *
     * These are only used for the electroweak part. In the QCD part
     * pole resummation is performed at the level of the polarisation function.
     */
    template<int N>
    inline
    complex Green_function_PR(complex lam, parameters const & p){
      complex result = Green_function<N>(lam,p);
      for(int n = 1; n <= p.resum_poles; ++n){
	result += resum_pole_GF<N>(n, p);
      }
      return result;
    }

    //! Green function corrections with pole resummation and weak scale in real part
    /**
     * @tparam N      Perturbative order. Leading order is N = 0.
     * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p       Further parameters.
     * @returns       The Nth order correction to the pole-resummed Green function. For the real part, the weak scale is used.
     *
     * Note that in Green_function<> the soft scale mu is used everywhere.
     * For the electroweak corrections, however, the real part is associated
     * with finite-width divergences.
     * For this reason the scale in the real part should be set to mu_weak,
     * as done in this function
     */
    template<int N>
    complex Green_function_PR_mu_weak_in_real(
       complex lam, parameters const & p
    );

    //! Corrections from insertion-scheme masses to the Green function
    /**
     * @tparam N      Perturbative order of the Green function before
     *                expansion around the leading-order mass.
     * @tparam DM     Order in expansion around leading-order mass.
     * @param lam     The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p       Further parameters.
     * @returns       The threshold-mass correction to the Green function.
     *
     * When using threshold (insertion) mass schemes
     * m_thr = m_Pole + delta_m_0 + delta_m_1 + ...
     * the Green functions should be expanded for small delta_m_1, delta_m2 ...
     *
     * Green_function_TM<N, DM> is the expansion term of order DM in the
     * expansion of Green_function<N>. Thus, the total perturbative order
     * is N + DM.
     *
     */
    template<int N, int DM>
    complex Green_function_TM(complex lam, parameters const & p);

    //! Counterterms to Green functions from weak corrections
    /**
     * @tparam N      Perturbative order.
     * @param p       Further parameters.
     * @returns       Corrections to the Green function from choosing mu_weak
     *                different from the overall renormalisation scale mu.
     */
    template<int N>
    double Green_function_weak(parameters const & p);

  }
}
