/**
 * @file
 * @brief Asymptotic expansions for harmonic sums
 *
 * Values of nested harmonic sums for large argument.
 *
 * This header is required by harmonic_sums.hpp.
 *
 * The formulas were generated with the asymptotic_sums FORM package.
 *
 * Reference: Albino, arXiv:0902.2148
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "complex.hpp"
#include "zeta.hpp"

namespace harmonic_sums{

  namespace detail{

    using namespace complex;
    using namespace zeta;

    constexpr double x_min=20.;

    //! Bernoulli numbers
    constexpr auto B = std::array<double,15>
      {{1, -1./2., 1./6., 0, -1./30., 0, 1./42., 0, -1./30., 0, 5./66., 0, -691./2730., 0, 7./6.}}
      ;

    //! asymptotic expansions for harmonic sums
    // it would be great to use simple functions here,
    // but partial template specialisation is not allowed then
    template<int N, int... indices>
    struct asymptotic_harmonic_sum{
      template<typename T>
      auto operator()(T z) -> decltype(1./z);
    };

    template<>
    struct asymptotic_harmonic_sum<1>{
      template<typename T>
      auto operator()(T z) -> decltype(1./z){
	typedef decltype(1./z) return_t;
	return_t const den_z = 1./z;
	return_t const den_z2 = den_z*den_z;
	// S_1(z) = gamma_Euler + log(z) - sum B_i/(i*z^i)
	return_t result = gamma_Euler + log(z) - B[1]*den_z;
	return_t den_zi = 1; //den_zi = 1/z^i
	for(size_t i=2; i < B.size(); i+=2){
	  den_zi *= den_z2;
	  result -= B[i]*den_zi/static_cast<int>(i);
	}
	return result;
      }
    };

    template<int N>
    struct  asymptotic_harmonic_sum<N>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	static_assert(N >= 2, "N too small");
	auto const den_z = 1./z;
	auto const den_z2 = den_z*den_z;
	auto den_zi = pow(den_z,N-1); //den_zi = 1/z^i
	// S_N(z) = zeta(n) - 1/(N-1)! sum (N+i-2)!/i! B_i/z^(i+N-1)
	double coeff = 1./(N-1);
	auto result = zeta::zeta(N) - B[1]*den_zi*den_z;
	for(size_t i = 0; i < B.size(); i+=2){
	  result -= coeff*B[i]*den_zi;
	  den_zi *= den_z2;
	  coeff *= (N+i)*(N+i-1.)/((i+1.)*(i+2.));
	}
	return result;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<2,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const l = log(z) + gamma_Euler;
	return x*(-1.+x*(-1./4.+x*(13./36.+x*(-1./6.+x*(-1./100.+x*(11./240.+x*(601./35280.+x*(-11./252.+x*(-247./7560.+x*(37./480.+x*(327053./3659040.+x*(-7./33.+(-95538931./281080800.+54589./65520.*x)*x))))))))))))+l*x*(-1.+x*(1./2.+x*(-1./6.+x2*(1./30.+x2*(-1./42.+x2*(1./30.+(-5./66.+691./2730.*x2)*x2))))))+2.*zeta_3;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<3,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const l = log(z) + gamma_Euler;
	return (-1./4.+x*(-1./6.+x*(17./48.+x*(-5./24.+x*(-23./720.+x*(7./80.+x*(629./10080.+x*(-19./168.+x*(-209./1400.+x*(121./480.+x*(65833./133056.+(-221./264.-134178917./60540480.*x)*x)))))))))))*x2+l*x2*(-1./2.+x*(1./2.+x*(-1./4.+x2*(1./12.+x2*(-1./12.+x2*(3./20.+x2*(-5./12.+691./420.*x2)))))))+5.*zeta_4/4.;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<3,2>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	return x2*(-zeta_2/2.+x*(1./3.+zeta_2/2.+x*(-5./8.-zeta_2/4.+x*(37./60.+x*(-7./24.+zeta_2/12.+x*(-37./420.+x*(13./80.-zeta_2/12.+x*(118./945.+x*(-11./42.+3.*zeta_2/20.+x*(-785./2772.+x*(169./240.-5.*zeta_2/12.+x*(42376./45045.+x*(-725./264.+691.*zeta_2/420.)))))))))))))+3.*zeta_2*zeta_3-9.*zeta_5/2.;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<4,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const x3 = x*x2;
	auto const l = log(z) + gamma_Euler;
	return (-1./9.+x*(-1./8.+x*(7./20.+x*(-1./4.+x*(-157./2520.+x*(3./20.+x*(173./1134.+x*(-127./504.+x*(-3769./8316.+x*(331./480.+(557347./308880.-179./66.*x)*x))))))))))*x3+l*(-1./3.+x*(1./2.+x*(-1./3.+x2*(1./6.+x2*(-2./9.+(1./2.-5./3.*x2)*x2)))))*x3-zeta_2*zeta_3+3.*zeta_5;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<4,2>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const x3 = x2*x;
	return x3*(x*(1./4.+x*(-3./5.+x*(25./36.+x*(-1./3.+x*(-143./720.+x*(1./4.+x*(101./280.+x*(-43./84.+x*(-5063./5040.+x*(5./3.+18329./4620.*x-5.*zeta_2/3.))+zeta_2/2.))-2.*zeta_2/9.))+zeta_2/6.))-zeta_2/3.)+zeta_2/2.)-zeta_2/3.)-zeta_6/3.+zeta_3*zeta_3;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<5,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const x4 = x2*x2;
	auto const l = log(z) + gamma_Euler;
	return (-1./16.+x*(-1./10.+x*(25./72.+x*(-7./24.+x*(-97./960.+x*(19./80.+x*(1037./3360.+x*(-253./504.+x*(-67751./60480.+x*(793./480.+703757./133056.*x))))))))))*x4+l*(-1./4.+x*(1./2.+x*(-5./12.+x2*(7./24.+x2*(-1./2.+(11./8.-65./12.*x2)*x2)))))*x4+7.*zeta_6/4.-zeta_3*zeta_3/2.;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<2,1,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const l = log(z) + gamma_Euler;
	return l*x*(-1.+x*(-1./4.+x*(13./36.+x*(-1./6.+x*(-1./100.+x*(11./240.+x*(601./35280.+x*(-11./252.+x*(-247./7560.+x*(37./480.+x*(327053./3659040.+x*(-7./33.+(-95538931./281080800.+54589./65520.*x)*x))))))))))))+l*l*x*(-1./2.+x*(1./4.+x*(-1./12.+x2*(1./60.+x2*(-1./84.+x2*(1./60.+(-5./132.+691./5460.*x2)*x2))))))+x*(-1.+x*(1./8.+x*(-79./216.+x*(37./96.+x*(-6587./36000.+x*(-23./960.+x*(20231./329280.+x*(403./10080.+x*(-120283./1814400.+x*(-107491./1209600.+x*(7459229./59628800.+x*(3349./11880.+x*(-71825424101./204626822400.-4458387211./3632428800.*x+(691.*zeta_2)/5460.))-(5.*zeta_2)/132.))+zeta_2/60.))-zeta_2/84.))+zeta_2/60.))-zeta_2/12.)+zeta_2/4.)-zeta_2/2.)+3.*zeta_4;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<2,2,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const l = log(z) + gamma_Euler;
	return l*(1./2.+x*(-2./3.+x*(13./24.+x*(-1./4.+x*(-7./360.+x*(1./10.+x*(23./1260.+x*(-31./252.+x*(-121./4200.+x*(4./15.+x*(481./6930.+(-115./132.-3015581./12612600.*x)*x)))))))))))*x2+x*(-2.*zeta_3+x*(3./4.+x*(-17./36.+x*(-11./288.+x*(61./180.+x*(-53./200.+x*(-67./3150.+x*(161009./1058400.+x*(7477./127008.+x*(-2570231./10584000.+x*(-267./1540.+x*(991776839./1536796800.+x*(68935./100672.-22725120853489./9090153072000.*x+(691.*zeta_3)/1365.))-(5.*zeta_3)/33.))+zeta_3/15.))-zeta_3/21.))+zeta_3/15.))-zeta_3/3.)+zeta_3))+2.*zeta_5;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<3,1,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const l = log(z) + gamma_Euler;
	return l*(-1./4.+x*(-1./6.+x*(17./48.+x*(-5./24.+x*(-23./720.+x*(7./80.+x*(629./10080.+x*(-19./168.+x*(-209./1400.+x*(121./480.+x*(65833./133056.+(-221./264.-134178917./60540480.*x)*x)))))))))))*x2+l*l*x2*(-1./4.+x*(1./4.+x*(-1./8.+x2*(1./24.+x2*(-1./24.+x2*(3./40.+(-5./24.+691./840.*x2)*x2))))))+x2*(-1./8.+x*(1./9.+x*(-65./192.+x*(101./240.+x*(-1813./8640.+x*(-1481./20160.+x*(743./8064.+x*(5021./36288.+x*(-692203./6048000.+x*(-4896769./13305600.+x*(17432497./79833600.+x*(7912873./5765760.+x*(-5243133011./10170800640.+(691.*zeta_2)/840.))-(5.*zeta_2)/24.))+(3.*zeta_2)/40.))-zeta_2/24.))+zeta_2/24.))-zeta_2/8.)+zeta_2/4.)-zeta_2/4.)+zeta_2*zeta_3-zeta_5/2.;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<4,1,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const x3 = x2*x;
	auto const l = log(z) + gamma_Euler;
	return l*(-1./9.+x*(-1./8.+x*(7./20.+x*(-1./4.+x*(-157./2520.+x*(3./20.+x*(173./1134.+x*(-127./504.+x*(-3769./8316.+x*(331./480.+(557347./308880.-179./66.*x)*x))))))))))*x3+l*l*(-1./6.+x*(1./4.+x*(-1./6.+x2*(1./12.+x2*(-1./9.+(1./4.-5./6.*x2)*x2)))))*x3+x3*(-1./27.+x*(3./32.+x*(-193./600.+x*(11./24.+x*(-5569./23520.+x*(-853./5760.+x*(27019./204120.+x*(337./1008.+x*(-4438559./24393600.+x*(-47587./44800.+x*(203482541./674593920.+171053./36960.*x-(5.*zeta_2)/6.))+zeta_2/4.))-zeta_2/9.))+zeta_2/12.))-zeta_2/6.)+zeta_2/4.)-zeta_2/6.)+(89.*zeta_6)/48.-zeta_3*zeta_3/2.;
      }
    };

    template<>
    struct  asymptotic_harmonic_sum<2,1,1,1>{
         template<typename T>
      auto operator()(T z) -> decltype(1./z){
	auto const x = 1./z;
	auto const x2 = x*x;
	auto const l = log(z) + gamma_Euler;
	return l*l*x*(-1./2.+x*(-1./8.+x*(13./72.+x*(-1./12.+x*(-1./200.+x*(11./480.+x*(601./70560.+x*(-11./504.+x*(-247./15120.+x*(37./960.+x*(327053./7318080.+x*(-7./66.+(-95538931./562161600.+54589./131040.*x)*x))))))))))))+l*l*l*x*(-1./6.+x*(1./12.+x*(-1./36.+x2*(1./180.+x2*(-1./252.+x2*(1./180.+(-5./396.+691./16380.*x2)*x2))))))+l*x*(-1.+x*(1./8.+x*(-79./216.+x*(37./96.+x*(-6587./36000.+x*(-23./960.+x*(20231./329280.+x*(403./10080.+x*(-120283./1814400.+x*(-107491./1209600.+x*(7459229./59628800.+x*(3349./11880.+x*(-71825424101./204626822400.-4458387211./3632428800.*x+(691.*zeta_2)/5460.))-(5.*zeta_2)/132.))+zeta_2/60.))-zeta_2/84.))+zeta_2/60.))-zeta_2/12.)+zeta_2/4.)-zeta_2/2.)+x*(-1.-zeta_2/2.+x*(1./16.-zeta_2/8.+x*(65./648.+(13.*zeta_2)/72.+x*(-127./384.-zeta_2/12.+x*(33919./90000.-zeta_2/200.+x*(-679./3456.+(11.*zeta_2)/480.+x*(-3670669./124467840.+(601.*zeta_2)/70560.+x*(27527./362880.-(11.*zeta_2)/504.+x*(497429./8164800.-(247.*zeta_2)/15120.+x*(-135439./1512000.+(37.*zeta_2)/960.+x*(-5080026913./31877556480.+(327053.*zeta_2)/7318080.+x*(131887./760320.-(7.*zeta_2)/66.+x*(8616607154069./14963336388000.+x*(-462772057751./1017080064000.+(54589.*zeta_2)/131040.)-(95538931.*zeta_2)/562161600.+(691.*zeta_3)/8190.))-(5.*zeta_3)/198.))+zeta_3/90.))-zeta_3/126.))+zeta_3/90.))-zeta_3/18.)+zeta_3/6.)-zeta_3/3.)+4.*zeta_5;
      }
    };

  }
}
