/**
 * @file
 * @brief Nested harmonic sums
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <array>
#include <string>
#include <cmath>
#include <complex>

#include "complex.hpp"

#include "asymptotic_harmonic_sums.hpp"

namespace harmonic_sums{

  using namespace complex;
  typedef std::complex<double> complex;

  namespace detail{
    // constexpr alternative to std::abs(int)
    inline constexpr int abs(int n){
      return (n>0)?n:-n;
    }

    inline constexpr bool is_positive(int n){
      return n > 0;
    }

    inline constexpr bool is_negative(int n){
      return n < 0;
    }

    /** \internal
     *  pow function optimised for positive integer exponents
     *
     *  @tparam N   Exponent
     *  @param t    Base
     *  @returns    t to the power N
     */
    template<int N, typename T>
    inline
    typename std::enable_if<is_positive(N), T>::type
    pow(T t){
      T result = 1;
      T prod = t;
      for(int i = 1; i <= N; i *= 2){
	if(N & i) result *= prod;
	prod *= prod;
      }
      return result;
    }

    /** \internal
     *  pow function optimised for negative integer exponents
     *
     *  @tparam N   Exponent
     *  @param t    Base
     *  @returns    t to the power N
     */
    template<int N, typename T>
    inline
    typename std::enable_if<is_negative(N), T>::type
    pow(T t){
      return 1./pow<-N>(t);
    }
    using std::pow;

    /** \internal
     * Implementation of nested harmonic sum
     */
    template<int N, int... indices>
    class harmonic_sum_impl{

    public:
      /* \internal
       * Analytic continuation of nested harmonic sum to complex argument
       */
      complex operator()(complex z){
	//for large arguments use asymptotic result
	if(z.real() > x_min){
	  return asymptotic_harmonic_sum<N, indices...>{}(z);
	}

	// shift argument
	int shift =
	  1
	  + static_cast<int>(std::ceil(x_min))
	  - static_cast<int>(std::floor(z.real()))
	  ;
	complex result = asymptotic_harmonic_sum<N, indices...>{}(z+shift);
	auto S_indices = harmonic_sum_impl<indices...>{};
	for(int m = 1; m <= shift; ++m){
	  //remove comment to make this work for negative indices?
	  result -= /* pow(copysign(1.0,N),m+z) * */
	    pow<-abs(N)>(m+z) * S_indices(z+m);
	}
	return result;
      }

      /* \internal
       * Nested harmonic sum for integer argument
       */
      double operator()(int n){
	//for large arguments use asymptotic result
	// if(n > x_min){
	//   return asymptotic_harmonic_sum<N, indices...>{}(n);
	// }

	//sum up explicitely
	double result = 0;
	auto S_indices = harmonic_sum_impl<indices...>{};
	for(int i = 1; i <= n; ++i){
	  result += S_indices(i)/pow(i,N)/*pow<-N>(i)*/;
	}

	return result;
      }
    };

    /** \internal
     * Implementation of single harmonic sum
     */
    template<int N>
    class harmonic_sum_impl<N>{
    public:
      /* \internal
       * Analytic continuation of single harmonic sum to complex argument
       */
      complex operator()(complex z){
	complex result = 0;
	while(z.real() < x_min){
	  z += 1;
	  result -= pow<-N>(z);
	}
	return result + asymptotic_harmonic_sum<N>{}(z);
      }

      /* \internal
       * Single harmonic sum for integer argument
       */
      double operator()(int n){
	if(n > x_min) return asymptotic_harmonic_sum<N>{}(n);
	double result = 0;
	for(int i = 1; i <= n; ++i){
	  result += pow(i, -N)/*pow<-N>(i)*/;
	}
	return result;
      }
    };

    /** \internal
     * Implementation of double harmonic sums
     * exploit relation from shuffle algebra
     */
    template<int N>
    class harmonic_sum_impl<N, N>{
    public:
      complex operator()(complex z){
	const auto SN = harmonic_sum_impl<N>{}(z);
	return (SN*SN + harmonic_sum_impl<2*abs(N)>{}(z))/2.;
      }

      double operator()(int n){
	const auto SN = harmonic_sum_impl<N>{}(n);
	return (SN*SN + harmonic_sum_impl<2*abs(N)>{}(n))/2.;
      }
    };

    //equations in Appendix A of arXiv:0901.3106
    template<>
    class harmonic_sum_impl<1, 1, 1>{
    public:
      complex operator()(complex z){
	const auto S1 = harmonic_sum_impl<1>{}(z);
	return (
		+ S1*S1*S1
		+ 3*S1*harmonic_sum_impl<2>{}(z)
		+ 2*harmonic_sum_impl<3>{}(z)
		)/6.;
      }

      double operator()(int n){
	const auto S1 = harmonic_sum_impl<1>{}(n);
	return (
		+ S1*S1*S1
		+ 3*S1*harmonic_sum_impl<2>{}(n)
		+ 2*harmonic_sum_impl<3>{}(n)
		)/6.;
      }
    };
  }

  //! Nested harmonic sum
  /**
   * harmonic_sum<a>(z) is the harmonic sum \f$\sum_{i=1}^z 1/i^a1 =f$
   *
   * harmonic_sum<a1,...,am>(z) is the nested harmonic sum
   * \f$\sum_{i=1}^z {\tt harmonic_sum<a2,...,am>(z)}/i^a1\f$
   */
  template<int a1, int... indices>
  double harmonic_sum(int n){
    return detail::harmonic_sum_impl<a1, indices...>{}(n);
  }

  //! Nested harmonic sum
  /**
   * harmonic_sum<a>(z) is the analytic continuation of the
   * harmonic sum \f$\sum_{i=1}^z 1/i^a1 =f$ to real values of z
   *
   * harmonic_sum<a1,...,am>(z) is the analytic continuation of the nested
   * harmonic sum \f$\sum_{i=1}^z {\tt harmonic_sum<a2,...,am>(z)}/i^a1\f$ to
   * real values of z
   */
  template<int a1, int... indices>
  complex harmonic_sum(double z){
    return detail::harmonic_sum_impl<a1, indices...>{}({z,0});
  }

  //! Nested harmonic sum
  /**
   * harmonic_sum<a>(z) is the analytic continuation of the
   * harmonic sum \f$\sum_{i=1}^z 1/i^a1 =f$ to complex values of z
   *
   * harmonic_sum<a1,...,am>(z) is the analytic continuation of the nested
   * harmonic sum \f$\sum_{i=1}^z {\tt harmonic_sum<a2,...,am>(z)}/i^a1\f$ to
   * complex values of z
   */
  template<int a1, int... indices>
  complex harmonic_sum(complex z){
    return detail::harmonic_sum_impl<a1, indices...>{}(z);
  }

}
