/**
 * @file
 * @brief Integration of real-valued functions
 *
 * This header provides a c++ interface to gsl quadpack integration
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
/*  */

#include <new>

#include "gsl/gsl_integration.h"

namespace integrate{

  namespace detail{
    /** \internal
     * Provides a function with the signature required by gsl
     * that can actually call an arbitrary Functor
     */
    template<class Functor>
    inline
    double call_from_gsl(double t, void * p){
      Functor * f = static_cast<Functor *>(p);
      return (*f)(t);
    }
  }

  //! Class for integrating a function of type Integrand.
  /**
   * The function has to be real-valued.
   */
  template<class Integrand>
  class integrator{
    static constexpr size_t workspace_size = 200;
    Integrand i;
    gsl_integration_cquad_workspace * space;
    double abserr;
    size_t nevals;
  public:
    //! Constructor
    /**
     * @param in    The integrand function.
     */
    integrator(Integrand in):
      i(in),
      space(gsl_integration_cquad_workspace_alloc(workspace_size)){
      if(space == nullptr) throw std::bad_alloc();
    }
    ~integrator(){
      gsl_integration_cquad_workspace_free(space);
    }
    //! Integrate over the integrand specified during construction.
    /**
     * @param a       Lower bound of the integration.
     * @param b       Upper bound of the integration.
     * @param epsabs  Absolute accuracy
     * @param epsrel  Relative accuracy
     * @returns       The integral from a to b with the specified accuracy.
     */
    double integrate(
	double a, double b,
	double epsabs = 10e-8, double epsrel = 10e-6
    ){
      double result;
      gsl_function F;
      F.function = &detail::call_from_gsl<Integrand>;
      F.params = &i;
      gsl_integration_cquad(
	  &F, a, b, epsabs, epsrel, space, &result, &abserr, &nevals
      );
      return result;
    }
    //! Obtain absolute error of last integration
    double get_error() const {
      return abserr;
    }
    //! Obtain number of evaluations during last integration
    double get_num_evals() const {
      return nevals;
    }
  };

  //! helper function to construct an integrator object
  /**
   * @param I   The integrand
   * @returns   An integrator object for I
   */
  template<class Integrand>
  inline
  integrator<Integrand> make_integrator(Integrand I){
    return integrator<Integrand>{I};
  }
}
