/**
 * @file
 * @brief Functions to load precomputed grids
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <string>
#include <fstream>

//!-
namespace QQbar_threshold{
  //! Loads a precomputed grid
  /**
   *
   * @param filename  Name of the grid to be loaded
   *
   * For order > 1, the cross section functions require a precomputed
   * grid. Some grids are already packaged with this library. New grids
   * can be generated with the QQbarGridCalc.m Mathematica package
   *
   * @see bbbar_xsection, ttbar_xsection, QQbarGridCalc.m, \ref QQbarThreshold
   **/
  void load_grid(char const* filename);
  //! @overload
  void load_grid(std::string const & filename);
  //! @overload
  /**
   * @param in  stream to the grid to be loaded
   */
  void load_grid(std::ifstream && in);

  //! Loads a precomputed grid for the non-resonant contribution
  /**
   *
   * @param filename  Name of the grid to be loaded
   *
   * Explicitly loading a grid is only required for exotic values
   * of the top mass or the Z mass. For these cases grids
   * can be generated with the QQbarGridCalc.m Mathematica package
   *
   **/
  void load_nonresonant_grid(char const* filename);
  //! @overload
  void load_nonresonant_grid(std::string const & filename);
  //! @overload
  /**
   * @param in  stream to the grid to be loaded
   */
  void load_nonresonant_grid(std::ifstream && in);

  //! Get directory with default grids
  /**
   *  @returns  The directory containing the default grids
   */
  std::string grid_directory();

  //! Grid range in rescaled energy and width coordinates
  struct Et_Gammat_range{
    //! Minimum and maximum rescales energy Et = - 4/(alpha_s^2*c_F^2*m)*E
    double Et_min, Et_max;
    //! Minimum and maximum rescales width Gammat = - 4/(alpha_s^2*c_F^2*m)*Gamma
    double Gammat_min, Gammat_max;
  };

  //! Get grid range
  /**
   *  @returns  The coordinate range of the currently loaded grid.
   *
   *  @see Et_Gammat_range
   */
  Et_Gammat_range grid_range();
}
