/**
 * @file
 * @brief Polygamma function
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "complex.hpp"
#include "zeta.hpp"
#include "harmonic_sums.hpp"

namespace psi{
  using namespace complex;
  using namespace zeta;
  using zeta::zeta;
  using harmonic_sums::harmonic_sum;
  typedef std::complex<double> complex;


  namespace detail{
    constexpr unsigned long long fac(unsigned n){
      return (n==0)?1:n*fac(n-1);
    }
  }

  //! The polygamma function for complex argument
  /**
   * @tparam N  Order of the polygamma function
   * @param z   Complex argument
   * @returns   \f$\psi^{(N)}(z)$\f
   */
  template<int N>
  complex psi(complex z){
    return ((N % 2)?-1:+1) * (detail::fac(N) * (harmonic_sum<N+1>(z-1.) - zeta(N+1)));
  }

  //! The polygamma function for integer argument
  /**
   * @tparam N  Order of the polygamma function
   * @param n   integer argument
   * @returns   \f$\psi^{(N)}(n)$\f
   */
  template<int N>
  double psi(int n){
    return ((N % 2)?-1:+1) * (detail::fac(N) * (harmonic_sum<N+1>(n-1) - zeta(N+1)));
  }

  template<>
  inline complex psi<0>(complex z){
    return (harmonic_sum<1>(z-1.) - gamma_Euler);
  }

  template<>
  inline double psi<0>(int n){
    return (harmonic_sum<1>(n-1) - gamma_Euler);
  }

}
