/**
 * @file
 * @brief Functions to calculate the residues of \f$q\bar{q}\f$ bound states
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include "parameters.hpp"

/// -
namespace QQbar_threshold{

  //! Computes the residue \f$Z_n\f$ of an S-wave \f$t\bar{t}\f$ bound state
  /**
   *
   * @param n     principal quantum number
   * @param mu    renormalisation scale
   * @param m     top quark mass
   * @param o     perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt   further options
   * @return      The residue \f$Z_n\f$.
   *
   * The exact definition of the residue depends on the setting of the option
   * \ref options::expand_s "expand_s". If set to \c true, it is given by
   * \f[Z_N = \mathop{\rm Res}_{E = E_N}\frac{2}{3}m^2 \Pi(E)\,,\f] otherwise by
   * \f[Z_N = \mathop{\rm Res}_{E = E_N}\frac{s}{6}\Pi(E)\,.\f] In both cases
   * \f$\Pi\f$ is the polarisation function in the full theory
   * including the quark electric charge.
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::expand_s = false
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   *
   * @see options, ttbar_energy_level, bbbar_residue
   */
  double ttbar_residue(
      int n, double mu, double m, order o,
      options opt = top_options()
  );

  //! Computes the residue \f$Z_n\f$ of an S-wave \f$b\bar{b}\f$ bound state
  /**
   *
   * @param n     principal quantum number
   * @param mu    renormalisation scale
   * @param m     bottom quark mass
   * @param o     perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt   further options
   * @return      The binding energy \f$E_n\f$.
   *
   * The exact definition of the residue depends on the setting of the option
   * \ref options::expand_s "expand_s". If set to \c true, it is given by
   * \f[Z_N = \mathop{\rm Res}_{E = E_N}\frac{2}{3}m^2 \Pi(E)\,,\f] otherwise by
   * \f[Z_N = \mathop{\rm Res}_{E = E_N}\frac{s}{6}\Pi(E)\,.\f] In both cases
   * \f$\Pi\f$ is the polarisation function in the full theory
   * including the quark electric charge.
   *
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_bottom}
   *   - options::expand_s = false
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::QED
   *   - options::alpha = #alpha_Y
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 1220.3 (for nl == 4)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   *
   * @see options, bbbar_energy_level, ttbar_residue
   */
  double bbbar_residue(
      int n, double mu, double m, order o,
      options opt = bottom_options()
  );

  namespace detail{
    /** \internal
     * Main function for the residue.
     * @param n    Principal quantum number.
     * @param p    Further parameters.
     * @return     The residue.
     */
    double QQbar_residue(int n, parameters const & p);

    /** \internal
     * Leading-order residue.
     * @param n    Principal quantum number.
     * @param p    Further parameters.
     * @return     The residue.
     */
    double Z_LO(int n, parameters const & p);

    // higher order corrections
    // Z = Z_LO * (1 + z1 + ... + z<order>)
    double z1(int n, parameters const & p);
    double z2(int n, parameters const & p);
    double z3(int n, parameters const & p);

    //NLO corrections - needed for double insertion
    double delta_f1_light(int n, double xi_n);
    double delta_f1_Coulomb(int n, parameters const & p);
    //needed for pole resummation in the Green function
    double f1(int n, parameters const & p);
  }

}
