/**
 * @file
 * @brief Implementation of interpolation for sums provided by grids.
 *
 * @see insertion_sums.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <complex>
#include <algorithm>
#include <string>
#include <exception>
#include <stdexcept>
#include <cassert>

#include "insertion_sums.hpp"
#include "interpolation_grid.hpp"
#include "parameters.hpp"
#include "QCD/colour.hpp"
#include "zeta.hpp"

namespace QQbar_threshold{
  namespace detail{
    typedef std::complex<double> complex;
    using zeta::pi;

    extern std::array<
      interpolation_grid<complex>,
      last_insertion_sum + 1
      > insertion_grids;

    //! Generate message for points outside the grid
    /**
     * @param what      Name of the coordinate that is out of bounds
     * @param point     Value of the out-of-bound coordinate
     * @param grid_min  Mininum value supported by the grid
     * @param grid_max  Maximum value supported by the grid
     * @returns         An error message
     */
    std::string out_of_grid_message(
	std::string const & what,
	double point,
	double grid_min,
	double grid_max
    ){
      int const exp = std::trunc(std::log10(std::abs(point)));
      return  "loaded grid too small:\n"
	"rescaled " + what + " "
	+ std::to_string(point*pow(10,-exp))
	+ "*10^" + std::to_string(exp) +
	+ " not in range ("
	+ std::to_string(grid_min*pow(10,-exp))
	+ "*10^" + std::to_string(exp) +
	+ ", "
	+ std::to_string(grid_max*pow(10,-exp))
	+ "*10^" + std::to_string(exp) +
	+ ")"
	;
    }

    //! Throw message if coordinate lies outside range
    /**
     * @param what      Name of the coordinate
     * @param point     Value of the coordinate
     * @param grid_min  Mininum value supported by the grid
     * @param grid_max  Maximum value supported by the grid
     *
     * Throws a std::out_of_range message if point is not
     * in the interval (grid_min, grid_max]
     */
    void throw_unless_in_grid(
	std::string const & what,
	double point,
	double grid_min,
	double grid_max
    ){
      if((point <= grid_min) || (point > grid_max)){
	throw std::out_of_range(
	    out_of_grid_message(
		what,
		point,
                grid_min,
                grid_max
	    ));
      }
    }

    //! Get interpolated value for an insertion sum
    /**
     * @param lam             The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param ins             Identifier of insertion sum.
     * @returns               The interpolated value of the insertion sum.
     */
    complex get_interpolation(complex lam, insertion_sum ins){
      using detail::insertion_grids;
      assert(ins < insertion_grids.size());
      if(insertion_grids[ins].empty()){
	throw std::logic_error("no grid (or empty grid) loaded");
      }
      complex invlam2 = pow(lam,-2);
      double e = real(invlam2); //rescaled energy
      throw_unless_in_grid(
          "energy", e,
          insertion_grids[ins].y_min(), insertion_grids[ins].y_max()
      );
      double g = imag(invlam2); // rescaled width
      throw_unless_in_grid(
          "width", g,
          insertion_grids[ins].x_min(), insertion_grids[ins].x_max()
      );
      return insertion_grids[ins](g, e);
    }

    struct en_point{
      double en;
      double val;
    };

    bool operator<(double a, en_point const &  b){
      return a < b.en;
    }

    //! ultrasoft insertion for zero width
    constexpr auto insertion_US_zero_width = std::array<en_point, 57>{{
        {0, 979},
        {0.05, 973.5224740913739},
        {0.1, 968.7001310250589},
        {0.15000000000000002, 963.6772964143879},
        {0.2, 958.6385208913299},
        {0.25, 953.4441828923578},
        {0.30000000000000004, 948.2903069428114},
        {0.35, 943.2090438609247},
        {0.39999999999999997, 938.3474715446204},
        {0.44999999999999996, 933.4820017073478},
        {0.5, 928.7334845914065},
        {0.55, 924.1463547145836},
        {0.6, 919.7381863558072},
        {0.7, 911.5168672805323},
        {0.7999999999999999, 903.8234527155706},
        {0.8999999999999999, 896.8713068026431},
        {1., 890.509554956274},
        {1.1, 884.7404136015732},
        {1.2, 879.5125288364051},
        {1.3, 874.7175715712442},
        {1.4, 870.3909112009338},
        {1.5, 866.458587915293},
        {1.6, 862.9982322644221},
        {1.7, 859.8593565111888},
        {1.9, 854.6087884966337},
        {2.1, 850.476786418282},
        {2.3, 847.4795254002913},
        {2.5, 844.9641981668548},
        {2.8, 843.3599389862707},
        {3.0999999999999996, 843.5722456610138},
        {3.4, 845.0005918027427},
        {3.6999999999999997, 847.9745445987596},
        {4., 852.5924509111026},
        {4.3, 857.7986673177085},
        {4.6, 864.0281340527505},
        {4.8999999999999995, 871.301525837364},
        {5.2, 879.1887132392021},
        {5.5, 888.0643584342041},
        {5.8, 897.7854045288772},
        {6.1, 908.1532992438536},
        {6.4, 919.3369982453111},
        {6.8, 935.698962015374},
        {7.2, 952.2549177137464},
        {7.6, 970.5182356696139},
        {8., 990.0218061249236},
        {9, 1042.4963638041716},
        {10, 1101.126829715978},
        {11, 1165.2123334220016},
        {12, 1234.77482343332},
        {13, 1309.1013334639715},
        {14, 1388.317109392158},
        {15, 1472.0719969773343},
        {16, 1560.2286114536832},
        {17, 1652.6406136574908},
        {18, 1749.28642514383},
        {19, 1849.8891075874105},
        {20, 1954.4222682947764}
      }};

    //! Get interpolated value for the ultrasoft constant for zero width
    /**
     * @param lam      The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @returns        The interpolated value of the ultrasoft constant.
     */
    complex get_us_const_zero_width(complex lam){
      double rescaled_energy = - real(pow(QCD::c_F/(2.*lam),2));
      auto const & grid = insertion_US_zero_width;
      auto upper =
	std::upper_bound(
	    grid.begin(),
	    grid.end(),
	    rescaled_energy
	);
      if((upper == grid.begin()) || (upper == grid.end())){
	throw std::out_of_range(
	    out_of_grid_message(
		"energy",
		rescaled_energy,
		grid.back().en,
		grid.front().en
	    ));
      }
      auto lower = upper - 1;
      //use linear interpolation between grid values
      return {0,
	  lower->val
	  + (rescaled_energy - lower->en)/(upper->en - lower->en)
	  *(upper->val - lower->val)
	  };

    }

    //! Get interpolated value for the ultrasoft constant
    /**
     * @param lam      The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @returns        The interpolated value of the ultrasoft constant.
     *
     * Depending on the width, uses either the loaded grid or the hard-coded
     * ultrasoft contribution for negligible width
     *
     * @see insertion_sums.hpp, insertion_matrices.hpp
     */
    complex insertion_US(complex lam){
      const double rescaled_width = - imag(1./pow(lam,2));
      assert(rescaled_width > 0);
      if(rescaled_width > 1e-4){
	return get_interpolation(lam, insertion_US_top);
      }
      return 2/(9.*pi*pi)*get_us_const_zero_width(lam);
    }

  }
}
