/**
 * @file
 * @brief Implementation of bound-state residues
 *
 * @see residues.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "residues.hpp"

#include <functional>

#include "energy_levels.hpp"
#include "scheme_conversion.hpp"
#include "harmonic_sums.hpp"
#include "zeta.hpp"
#include "QCD/colour.hpp"
#include "QCD/beta.hpp"
#include "alpha_s.hpp"
#include "Gamma.hpp"
#include "psi.hpp"
#include "integrate.hpp"
#include "Wilson_coefficients.hpp"
#include "log_n.hpp"
#include "h.hpp"

namespace QQbar_threshold{

  namespace detail{
    double QQbar_residue(int n, parameters const & p);
  }

  double ttbar_residue(
      int n, double mu, double m, order o,
      options opt
  ){
    return detail::QQbar_residue(n, gen_top_parameters(mu, m, o, opt));
  }

  double bbbar_residue(
      int n, double mu, double m, order o,
      options opt
  ){
    return detail::QQbar_residue(n, gen_bottom_parameters(mu, m, o, opt));
  }


  namespace detail{
    using namespace zeta;
    using QCD::beta;
    using QCD::c_F;
    using QCD::c_A;
    using QCD::T_f;
    using harmonic_sums::harmonic_sum;
    using namespace complex;
    using Gamma::Gamma;
    using psi::psi;

    // leading order residue
    double Z_LO(int n, parameters const & p){
      return pow((p.m*c_F*p.alpha_s)/n,3)/(8*pi);
    }

    /*************************************************************************/
    /* f1,f2,f3: higher-order corrections to the wave function at the origin.*/
    /* In PS-scheme take care of corrections from mass renormalisation.      */
    /* If the 1/s factor in the Green function is not expanded,              */
    /* the corrections are due to the m^3 factor from the non-relativistic   */
    /* wave function.                                                        */
    /* If the 1/s factor is expanded, there is an additional factor of m^-2  */
    /*************************************************************************/

    double f1(int n, parameters const & p);
    double f2(int n, parameters const & p);
    double f3(int n, parameters const & p);

    namespace{
      //! Insertion of massive light-quark correction Coulomb potential
      double delta_f1_light(int n, parameters const & p);
    }
    //! Insertion of NLO Coulomb potential
    /* not in anonymous namespace because it's needed by energy levels */
    double delta_f1_Coulomb(int n, parameters const & p);

    double f1(int n, parameters const & p){
      return p.contributions.v_Coulomb[0]*p.alpha_s/pi*(
	  + delta_f1_Coulomb(n, p)
	  + delta_f1_light(n, p)
      );
    }

    namespace{
      //! Insertion of NNLO Coulomb potential
      double delta_f2_double_Coulomb(int n, parameters const & p);
      //! Double insertion of NLO Coulomb potential
      double delta_f2_single_Coulomb(int n, parameters const & p);
      //! Insertion of delta (1/m^2) potential
      double delta_f2_delta(int n, parameters const & p);
      //! Insertion of 1/r^2 (1/m) potential
      double delta_f2_r2inv(int n, parameters const & p);
      //! Insertion of momentum-dependent potential
      double delta_f2_p2(int n, parameters const & p);
      //! Insertion of correction to kinetic energy
      double delta_f2_kinetic(int n, parameters const & p);
      //! Insertion of massive light-quark correction Coulomb potential
      double delta_f2_light(int n, parameters const & p);
      //! Double insertion of massive light quark and "massless" Coulomb potential
      double delta_f2_Coulomb_light(int n, parameters const & p);
      //! Double insertion of massive light quark correction Coulomb potential
      double delta_f2_double_light(int n, parameters const & p);
    }

    double f2(int n, parameters const & p){
      double result =
	+ p.contributions.v_Coulomb[1]*(
	    delta_f2_single_Coulomb(n, p) + delta_f2_light(n, p)
	)
	+ pow(p.contributions.v_Coulomb[0], 2)*delta_f2_double_Coulomb(n, p)
	+ p.contributions.v_delta[0]*delta_f2_delta(n, p)
	+ p.contributions.v_r2inv[0]*delta_f2_r2inv(n, p)
	+ p.contributions.v_p2[0]*delta_f2_p2(n, p)
	+ p.contributions.v_kinetic[0]*delta_f2_kinetic(n, p)
	;
      if(p.double_light_insertion){
	result += pow(p.contributions.v_Coulomb[0], 2)*(
	    delta_f2_Coulomb_light(n, p) + delta_f2_double_light(n, p)
	);
      }
      if(is_threshold_mass(p.mass_scheme)){
	const int mass_power = p.expand_s?1:3;
	result += mass_power*delta_m(0, p)/p.m;
      }
      return result;
    }

    namespace{
      //! Insertion of NNNLO Coulomb potential
      double delta_f3_single_Coulomb(int n, parameters const & p);
      //! Double insertion of NNLO and NLO Coulomb potential
      double delta_f3_double_Coulomb(int n, parameters const & p);
      //! Triple insertion of NLO Coulomb potential
      double delta_f3_triple_Coulomb(int n, parameters const & p);
      //! Double insertion of NLO Coulomb and delta (1/m^2) potential
      double delta_f3_Coulomb_delta(int n, parameters const & p);
      //! Double insertion of NLO Coulomb and 1/r^2 (1/m) potential
      double delta_f3_Coulomb_r2inv(int n, parameters const & p);
      //! Double insertion of NLO Coulomb and momentum-dependent potential
      double delta_f3_Coulomb_p2(int n, parameters const & p);
      //! Double insertion of NLO Coulomb potential and kinetic energy correction
      double delta_f3_Coulomb_kinetic(int n, parameters const & p);
      //! Insertion of correction to delta (1/m^2) potential
      double delta_f3_delta(int n, parameters const & p);
      //! Insertion of correction to 1/r^2 (1/m) potential
      double delta_f3_r2inv(int n, parameters const & p);
      //! Insertion of correction to momentum-dependent potential
      double delta_f3_p2(int n, parameters const & p);
      //! Insertion of Higgs potential
      double delta_f3_Higgs(int n, parameters const & p);
      //! Ultrasoft correction
      double delta_f3_ultrasoft(int n, parameters const & p);
    }

    double f3(int n, parameters const & p){
      double result =
	+ p.contributions.v_Coulomb[2] * delta_f3_single_Coulomb(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_Coulomb[1] * delta_f3_double_Coulomb(n, p)
	+ pow(p.contributions.v_Coulomb[0], 3) * delta_f3_triple_Coulomb(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_delta[0] * delta_f3_Coulomb_delta(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_r2inv[0] * delta_f3_Coulomb_r2inv(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_p2[0] * delta_f3_Coulomb_p2(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_kinetic[0] * delta_f3_Coulomb_kinetic(n, p)
	+ p.contributions.v_delta[1] * delta_f3_delta(n, p)
	+ p.contributions.v_r2inv[1] * delta_f3_r2inv(n, p)
	+ p.contributions.v_p2[1] * delta_f3_p2(n, p)
	+ p.contributions.v_Higgs[0] * delta_f3_Higgs(n, p)
	+ p.contributions.ultrasoft[0] * delta_f3_ultrasoft(n, p)
	;
      if(is_threshold_mass(p.mass_scheme)){
	const int mass_power = p.expand_s?1:3;
	result += (
	    + delta_m(1, p) * mass_power
	    + delta_m(0, p)*(
		+ mass_power * f1(n, p)
		- p.contributions.v_Coulomb[0]*3./2.*p.alpha_s/pi*4*beta(0, p.nl)
	    )
	)/p.m;
      }
      return result;
    }

    double z1(int n, parameters const & p){
      using Wilson_coefficients::cv;
      return 2*cv<1>(p) + f1(n, p);
    }

    double z2(int n, parameters const & p){
      using Wilson_coefficients::cv;
      using Wilson_coefficients::dv;
      return 2*cv<2>(p)
	+ pow(cv<1>(p),2)
	+ 2*cv<1>(p)*f1(n, p)
	+ f2(n, p)
	- E_LO(n, p)/p.m * (
	    + 1./3.*dv<0>(p)
	    + p.expand_s
	);
    }


    double z3(int n, parameters const & p){
      using Wilson_coefficients::cv;
      using Wilson_coefficients::dv;
      double result =
	+ 2*cv<3>(p)
	+ 2*cv<2>(p)*cv<1>(p)
	+ f1(n, p)*(2*cv<2>(p) + pow(cv<1>(p),2))
	+ 2*cv<1>(p)*f2(n, p)
	+ f3(n, p)
	- E_LO(n, p)/p.m*1./3.*dv<1>(p)
	;
      if(p.expand_s){
	result -= E_LO(n, p)/p.m*(
	    cv<1>(p)
	    + (1 + 1./3.*dv<0>(p))*(
		cv<1>(p) + e1_no_mass_shift(n, p) + f1(n, p)
	    )
	);
      }
      else{
	result -= E_LO(n, p)/p.m*1./3.*dv<0>(p)*(
	    cv<1>(p) + e1_no_mass_shift(n, p) + f1(n, p)
	);
      }
      return result;
    }

    double QQbar_residue(int n, parameters const & p){
      // Z = Z_LO * (1 + z<1> + ... + z<order>)
      double z_factor = 1;
      switch(p.order){
      case 3: z_factor += z3(n, p);
      case 2: z_factor += z2(n, p);
      case 1: z_factor += z1(n, p);
      default:;
      }
      if(p.expand_s) return z_factor * Z_LO(n, p);
      const double E_N = is_threshold_mass(p.mass_scheme)?
        QQbar_energy_level(n, p, p.m):
        QQbar_energy_level_no_shift(n, p)
        ;
      return pow(1 + E_N/(2*p.m), -2) * z_factor * Z_LO(n, p);
    }

      /***************************************************
       *        NLO corrections to wave function         *
       ***************************************************/

      double delta_f1_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return
	  + 3./2.*log_n(n, p)*4*beta(0,p.nl)
	  + 3./4.*a1_eff(p)
	  + 4*beta(0, p.nl)*(
	      + 1./2.*(harmonic_sum<1>(n) - 1)
	      + n*(harmonic_sum<2>(n) - zeta_2)
	  );
      }

    /************************* NLO light-quark correction *********************
     *                                                                        *
     * NLO massive light-quark correction to the residue.                     *
     * Equation numbers refer to                                              *
     * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132   *
     * \cite Beneke:2014pta						      *
     *									      *
     **************************************************************************/

    namespace{

      //! phi helper function
      /**
       * See eq. (C.13)
       * The sum over i (cf. eqs. (C.9), (C.10)) is already included here
       */
      complex phi_light(complex const & u, int n){
	assert(n > 0);
	/* possible optimisations:                                      */
	/* - use recursion of gammas                                    */
	/* - precompute u-independent part (typically, n will be small) */
	complex result = 0;
	for(int i = 1; i <= n; ++i){
	  result += Gamma(u+i)/
	    (Gamma(n-i+1)*pow(Gamma(i),2)*Gamma(i+1)*Gamma(u+2-i))*(
		+ (3-u)/n - psi<0>(n+1) + 2*psi<0>(i)
		+ psi<0>(1+i) - psi<0>(u+i) - psi<0>(u+2-i)
	    );
	}
	result *= Gamma(-u)*Gamma(u+2);
	return result;
      }

      //! Integrand of light-quark correction to e1
      /**
       * See eq. (C.9)
       */
      complex f1_light_integrand(int n, complex const & u, double xi){
	assert(n > 0);
	return phi_light(u, n)*Gamma(2-u/2)/(u*Gamma((5-u)/2)*pow(xi,u));
      }

      /* real part of the integrand after transforming the */
      /* integration variable to real values               */
      double re_f1_integrand(int n, double t, double xi){
	static constexpr double del = 0.5;
	auto u = complex{del, tan(pi/2*t)};
	return real(f1_light_integrand(n, u, xi))/pow(cos(pi/2*t),2);
      }

      //! Massive light-quark correction to f1
      /**
       * See eq. (C.9)
       *
       * f^{(1)}_{m_c} corresponds to alpha_s/pi*delta_f1_light
       */
      double delta_f1_light(int n, parameters const & p){
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	return detail::delta_f1_light(n, xi_n);
      }
    }

    double delta_f1_light(int n, double xi_n){
      using namespace std::placeholders;
      auto integrand = std::bind(re_f1_integrand, n, _1, xi_n);
      auto i = integrate::make_integrator(integrand);
      return -Gamma(n+1)*sqrt(pi)/16*i.integrate(-1, 1);
    }

    namespace{

      /***************************************************
       *       NNLO corrections to wave function         *
       ***************************************************/

      /************************* NNLO light-quark correction *********************
       *                                                                        *
       * NNLO massive light-quark correction to the residue.                    *
       * Equation numbers refer to                                              *
       * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132	*
       * \cite Beneke:2014pta							*
       *									*
       **************************************************************************/

      //! Integrand for NNLO light-quark correction to residue
      /**
       * The second part comes from using the MSbar scheme
       * for the light-quark mass.
       * See eq (C.17)
       */
      complex f2_light_integrand(int n, complex const & u, double xi, parameters const & p){
	assert(n > 0);
	return phi_light(u, n)*pow(xi,-u)*(
	    detail::light_integrand_NNLO(u, p) -
	    (2./3. + log(p.mu/p.ml))*Gamma(2 - u/2)/Gamma((5 - u)/2)
	);
      }

      /* real part of the integrand after transforming the */
      /* integration variable to real values               */
      double re_f2_integrand(int n, double t, double xi, parameters const & p){
	static constexpr double del = 0.5;
	auto u = complex{del, tan(pi/2*t)};
	return real(f2_light_integrand(n, u, xi, p))/pow(cos(pi/2*t),2);
      }

      //! Single massive light-quark correction to f2
      /**
       * See eqs. (C.10), (C.17)
       *
       * f^{(2)}_{m_c} corresponds to delta_f2_light (no extra factors here)
       */
      double delta_f2_light(int n, parameters const & p){
	using namespace std::placeholders;
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	auto integrand = std::bind(re_f2_integrand, n, _1, xi_n, p);
	auto i = integrate::make_integrator(integrand);
	return pow(p.alpha_s/pi,2)*Gamma(n+1)*sqrt(pi)/8*i.integrate(-1, 1);
      }

      /* charm mass correction to f2 - massive double insertion */
      static constexpr int s_max = 10;

      //! Double massive light-quark correction to f2
      /**
       * See eq. (C.19)
       *
       * f^{(2)}_{m_c, m_c} corresponds exactly to delta_f2_double_light
       *
       * @see h.hpp
       */
      double delta_f2_double_light(int n, parameters const & p){
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	const std::array<double, 3> hln = {
	  h_light<0>(n, n, xi_n),
	  h_light<1>(n, n, xi_n),
	  h_light<2>(n, n, xi_n)
	};
	double result =  (
	    + pow(hln[1],2)
	    + hln[0]*(
		+ 3*hln[0] - 4*hln[1]*(1 + n*psi<0>(n))
		+ hln[2] + n*hln[0]*(
		    4*psi<0>(n) + n*(2*zeta_2 + 2*pow(psi<0>(n),2) - psi<1>(n))
		)
	    )
	)/pow(n, 2);
	for(int s = 1; s < s_max; ++s){
	  if(s == n) continue;
	  result -= h_light<0>(n,  s, xi_n)*(
	      + 2*h_light<1>(n,  s, xi_n)
	      - h_light<0>(n,  s, xi_n)*(
		  2 + 2*n*psi<0>(n) + static_cast<double>(n)/(n-s)
	      )
	  )/(s*(s - n));
	}
	return  pi*pow(p.alpha_s/(4*pi*Gamma(n)),2)*result;
      }

      //! Mixed light mass/Coulomb correction to f2
      /**
       * See eq. (C.25)
       *
       * e^{(2)}_{m_c, massless} corresponds exactly to delta_e2_Coulomb_light
       *
       * @see h.hpp
       */
      double delta_f2_Coulomb_light(int n, parameters const & p){
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	double result = 0;
	for(int s = 1; s < s_max; ++s){
	  if(s == n) continue;
	  result += (
	      - h<-1>(n, s, p)*h_light<1>(n, s, xi_n)
	      - n*h<0>(n, s, p)*h_light<0>(n, s, xi_n)
	      + h<-1>(n, s, p)*h_light<0>(n, s, xi_n)*(
		  2 + n*psi<0>(n) + static_cast<double>(n)/(n-s)
	      )
	  )/(s*(n-s));
	}
	result += (
	    2*n*(
		- n*h<1>(n, n, p)*h_light<0>(n, n, xi_n)
		+ h<0>(n, n, p)*(
		    - h_light<1>(n, n, xi_n)
		    + h_light<0>(n, n, xi_n)*(2 + n*psi<0>(n))
		)
	    )
	    - h<-1>(n, n, p)*(
		+ h_light<2>(n, n, xi_n)
		- 2*h_light<1>(n, n, xi_n)*(2 + n*psi<0>(n))
		+ h_light<0>(n, n, xi_n)*(
		    6 + 4*n*psi<0>(n)
		    +n*n*(pow(psi<0>(n),2) - psi<1>(n) + 2*zeta_2)
		)
	    )
	)/(2*n*n);
	return  ((n%2)?1:-1)*pow(p.alpha_s/pi,2)*sqrt(pi)/(8*Gamma(n))*result;
      }


      /* correction to f2 from one insertion of coulomb potential */
      double delta_f2_single_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a;
	return pow(p.alpha_s/pi,2)/(48*n)*(
	    + 48*pow(n,2)*pow(4*beta(0,p.nl),2)*(harmonic_sum<3>(n) - harmonic_sum<2, 1>(n))
	    + 2*harmonic_sum<1>(n)*(
		+ 6*a<1>(p.nl)*n*4*beta(0,p.nl)
		+ pow(4*beta(0,p.nl),2)*(
		    - 36
		    - 12*n
		    + 18*n*harmonic_sum<1>(n)
		    + 24*pow(n,2)*(harmonic_sum<2>(n) - zeta_2)
		    + 12*n*log_n(n, p)
		)
		+ 3*n*16*beta(1,p.nl)
	    )
	    +n*(
		+ 9*a<2>(p.nl)
		- 12*a<1>(p.nl)*4*beta(0,p.nl)*(1 + 2*zeta_2*n)
		+ 54*zeta_2*pow(4*beta(0,p.nl),2)
		- 6*16*beta(1,p.nl)*(1 + 2*zeta_2*n)
		+ 2*(
		    + 18*a<1>(p.nl)*4*beta(0,p.nl)
		    - 12*(1 + 2*zeta_2*n)*pow(4*beta(0,p.nl),2)
		    + 9*16*beta(1,p.nl)
		)*log_n(n, p)
		+ 36*pow(4*beta(0,p.nl),2)*pow(log_n(n, p),2)
		+ 12*harmonic_sum<2>(n)*(
		    + 2*a<1>(p.nl)*n*4*beta(0,p.nl)
		    - pow(4*beta(0,p.nl),2) * (1 - 4*n*log_n(n, p))
		    + n*16*beta(1,p.nl)
		)
		+ 48*zeta_3*n*pow(4*beta(0,p.nl),2)
	    )
	);
      }

      /* correction to f2 from two insertions of coulomb potential */
      double delta_f2_double_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s/pi,2)*(
	    + 3./16.*pow(a1_eff(p),2)
	    + 4*beta(0, p.nl)*(
		a1_eff(p)*(
		    - 5./8.
		    - n*psi_1/2
		    + harmonic_sum<1>(n)/4
		    + 3./4.*log_n(n, p)
		)
		+ 4*beta(0, p.nl)*(
		    + 1./4.
		    - 3./4.*psi_1
		    + 3./4.*harmonic_sum<1>(n)/n
		    + harmonic_sum<1>(n)*(
			- 3./4.
			- harmonic_sum<1>(n)/4
			+ log_n(n, p)/2
		    )
		    + (
			- 5./4.
			+ 3./4.*log_n(n, p)
		    )*log_n(n, p)
		    + n*(
			- (harmonic_sum<2,1>(n) - 2*zeta_3)/2
			+ 3./4.*(harmonic_sum<3>(n) - zeta_3)
			+ n*(
			    - 5./4.*harmonic_sum<4>(n)
			    + harmonic_sum<3,1>(n)
			)
			- psi_1*(
			    - 1./2.
			    - n*psi_1/4
			    + log_n(n, p)
			)
		    )
		)
	    )
	);
      }

      /* correction to f2 from insertion of delta potential */
      double delta_f2_delta(int n, parameters const & p){
	using Wilson_coefficients::v0;
	using Wilson_coefficients::v0_ep;
	return pow(p.alpha_s,2)*pow(c_F,2)*(
	    2*(4 + n + 2*log_n(n, p)*n)*v0(p.nl)
	    + n*v0_ep(p.nl) - 4*n*v0(p.nl)*harmonic_sum<1>(n)
	)/(2*n);
      }

      /* correction to f2 from insertion of 1/r^2 potential */
      double delta_f2_r2inv(int n, parameters const & p){
	using Wilson_coefficients::b;
	using Wilson_coefficients::b_ep;
	return -pow(p.alpha_s,2)*c_F*(
	    + b_ep<1>(p.nl)*n
	    + b<1>(p.nl)*(
		8 + n*(6 - 2*log(2) + 4*log_n(n, p) - 4*harmonic_sum<1>(n))
	    )
	)/(4*n);
      }

      /* correction to f2 from insertion of momentum-dependent p-potential */
      double delta_f2_p2(int n, parameters const & p){
	return pow(p.alpha_s*c_F,2)*(
	    -5 + 16*n + 4*(3 + 2*log_n(n, p) - 2*harmonic_sum<1>(n))*pow(n,2)
	)/(4*pow(n,2));
      }

      /* correction to f2 from insertion of kinetic correction */
      double delta_f2_kinetic(int n, parameters const & p){
	return pow(p.alpha_s*c_F,2)*(
	    -5 + 8*n + (6 + 4*log_n(n, p) - 4*harmonic_sum<1>(n))*pow(n,2)
	)/(8*pow(n,2));
      }

      /***************************************************
       *       NNNLO corrections to wave function        *
       ***************************************************/

      /* correction to f3 from one insertions of coulomb potential */
      double delta_f3_single_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::a;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	double const S_3_diff =
	  + harmonic_sum<3>(n)
	  - harmonic_sum<2,1>(n)
	  + zeta_3
	  ;
	return pow(p.alpha_s/pi,3)*(
	    + 3./64.*a<3>(p.nl)
	    + (64*beta(2, p.nl) + 2*a<1>(p.nl)*16*beta(1, p.nl) + 48*zeta_2*pow(c_A,3))*(
		- 1./32.
		+ harmonic_sum<1>(n)/32
		+ 3./32.*log_n(n, p)
		- n*psi_1/16
	    )
	    +  4*beta(0, p.nl)*(
		+ a<2>(p.nl)*(
		    - 3./32.
		    + 3./32.*harmonic_sum<1>(n)
		    + 9./32.*log_n(n, p)
		    - 3./16.*n*psi_1
		)
		+ 16*beta(1, p.nl)*(
		    + 35./64.*zeta_2
		    + 5./32.*psi_1
		    +  harmonic_sum<1>(n)*(
			- 5./16.
			- 15./(16*n)
			+ 15./32.*harmonic_sum<1>(n)
			+ 5./16.*log_n(n, p)
			- 5./8.*n*psi_1
		    )
		    +  log_n(n, p)*(
			- 5./16.
			+ 15./32.*log_n(n, p)
			- 5./8.*n*psi_1
		    )
		    + 5./8.*n*S_3_diff
		)
		+ 4*beta(0, p.nl)*(
		    a<1>(p.nl)*(
			+ 21./32.*zeta_2
			+ 3./16.*psi_1
			+ harmonic_sum<1>(n)*(
			    -3./8.
			    - 9./(8*n)
			    + 9./16.*harmonic_sum<1>(n)
			    + 3./8.*log_n(n, p)
			    - 3./4.*n*psi_1
			)
			+ log_n(n, p)*(
			    -3./8.
			    + 9./16.*log_n(n, p)
			    - 3./4.*n*psi_1
			)
			+ 3./4.*n*S_3_diff
		    )
		    + 4*beta(0, p.nl)*(
			- 15./16.*zeta_2
			+ log_n(n, p)*(
			    + 21./16.*zeta_2
			    + (
				-3./8.
				+ 3./8.*log_n(n, p)
			    )*log_n(n, p)
			    + 3./8.*psi_1)
			+ 3./8.*psi_1
			+ harmonic_sum<1>(n)*(
			    + 3./(4*n)
			    + 39./16.*zeta_2
			    - 3./8.*psi_1
			    + harmonic_sum<1>(n)*(
				-3./8.
				+ harmonic_sum<1>(n)/8
				- 3./4.*n*psi_1
			    )
			    + log_n(n, p)*(
				-3./4.
				- 9./(4*n)
				+ 9./8.*harmonic_sum<1>(n)
				+ 3./8.*log_n(n, p)
				- 3./2.*n*psi_1
			    )
			    + 3./2.*n*S_3_diff
			)
			+ n*(
			    - 7./120.*pow(pi,4)
			    - 3*harmonic_sum<3, 1>(n)
			    + 3*harmonic_sum<2, 1, 1>(n)
			    - (
				- 9./8.*zeta_2
				+ 3./4.*pow(log_n(n, p),2)
				+ 3./4.*psi_1
			    )*psi_1
			    + 3./2.*log_n(n, p)*S_3_diff
			)
			+ harmonic_sum<3>(n)/4
			- 3./4.*harmonic_sum<2, 1>(n)
		    )
		)
	    )
	);
      }

      /* correction to f3 from two insertions of coulomb potential */
      double delta_f3_double_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::a;
	const double psi_1 = zeta_2 - harmonic_sum<2>(n);
        const double VC2_log_coeff =
          2*a<1>(p.nl)*4*beta(0, p.nl) + 16*beta(1, p.nl)
          ;
	return pow(p.alpha_s/pi,3)/n*(
            a1_eff(p)*n*(
                3./32.*a<2>(p.nl)
                + VC2_log_coeff*(
                    - 5./32.
                    + 3./16.*log_n(n, p)
                    - n*psi_1/8.
                    + harmonic_sum<1>(n)/16
                )
            )
            + 4*beta(0, p.nl)*(
                + 3./8.*VC2_log_coeff*harmonic_sum<1>(n)
                + 4*beta(0, p.nl)*(
                    - 3./4.*a1_eff(p)*harmonic_sum<1>(n)
                    + 4*beta(0, p.nl)*(
                        - 13./4.*zeta_2
                        + 13./4.*psi_1
                        + (3./4. - 13./4.*harmonic_sum<1>(n))*harmonic_sum<1>(n)
                        + 5*harmonic_sum<1, 1>(n)
                    )
                )
                + n*(
                    + a<2>(p.nl)*(
                        - 5./32.
                        + 3./16.*log_n(n, p)
                        - n*psi_1/8.
                        + harmonic_sum<1>(n)/16
                    )
                    + VC2_log_coeff*(
                        + 1./8.
                        - 3./8.*psi_1
                        + log_n(n, p)*(
                            - 5./8.
                            + 3./8.*log_n(n, p)
                            + harmonic_sum<1>(n)/4
                        )
                        + (-3./8. - harmonic_sum<1>(n)/8)*harmonic_sum<1>(n)
                        + n*(
                            + (1./4. - log_n(n, p)/2)*psi_1
                            + zeta_3/8.
                            + 3./8.*harmonic_sum<3>(n)
                            - harmonic_sum<2, 1>(n)/4
                            + n*(
                                + pow(psi_1, 2)/8
                                - 5./8.*harmonic_sum<4>(n)
                                + harmonic_sum<3, 1>(n)/2
                            )
                        )
                    )
                    + 4*beta(0, p.nl)*(
                        a1_eff(p)*(
                            + 1./8.
                            + 7./16.*zeta_2
                            + psi_1/8.
                            + log_n(n, p)*(
                                -5./8. + 3./8.*log_n(n, p)
                                + harmonic_sum<1>(n)/4
                            )
                            + (
                                -3./8. + 3./8.*harmonic_sum<1>(n)
                            )* harmonic_sum<1>(n)
                            + n*(
                                + zeta_3/2.
                                + psi_1*(
                                    + 1./4.
                                    - log_n(n, p)/2
                                    - harmonic_sum<1>(n)/2
                                )
                                + harmonic_sum<3>(n)/2
                                - harmonic_sum<2, 1>(n)/2
                            )
                        )
                        + 4*beta(0, p.nl)*(
                            - 17./16.*zeta_2
                            + 3./2.*zeta_3
                            + 5./8.*psi_1
                            + log_n(n, p)*(
                                + 3./4.
                                + 7./8.*zeta_2
                                - 5./4.*psi_1
                                + log_n(n, p)*(
                                    - 15./8.
                                    + 3./4.*log_n(n, p)
                                    + 3./4.*harmonic_sum<1>(n)
                                )
                                + (
                                    - 9./4. + harmonic_sum<1>(n)/4
                                )*harmonic_sum<1>(n)
                            )
                            + harmonic_sum<1>(n)*(
                                + 3./4.
                                - 3./8.*zeta_2
                                + 3./4.*psi_1
                                + (
                                    - 7./8. + harmonic_sum<1>(n)/4
                                )*harmonic_sum<1>(n)
                            )
                            + harmonic_sum<3>(n)/2
                            - 3./2.*harmonic_sum<2, 1>(n)
                            + n*(
                                - zeta_3
                                - 9./8.*zeta_4
                                + psi_1*(
                                    + 5./4.*zeta_2
                                    - psi_1/4.
                                    + log_n(n, p)*(
                                        + 3./2.
                                        - 3./2.*log_n(n, p)
                                        - harmonic_sum<1>(n)
                                    )
                                    + (
                                        + 3./2.
                                        - harmonic_sum<1>(n)/2
                                    )* harmonic_sum<1>(n)
                                )
                                + harmonic_sum<1>(n)* (
                                    zeta_3/2.
                                    - harmonic_sum<3>(n)/2
                                )
                                + 7./4.*harmonic_sum<4>(n)
                                + log_n(n, p)*(
                                    + 3./2.*zeta_3
                                    + 5./2.*harmonic_sum<3>(n)
                                    - 2*harmonic_sum<2, 1>(n)
                                )
                                + harmonic_sum<2, 1>(n)/2
                                - harmonic_sum<2, 2>(n)
                                - 3./2.*harmonic_sum<3, 1>(n)
                                + harmonic_sum<2, 1, 1>(n)
                                + n*(
                                    - zeta_2*zeta_3
                                    - zeta_5/2.
                                    - 2*harmonic_sum<5>(n)
                                    + psi_1*(
                                        - harmonic_sum<3>(n)
                                        + harmonic_sum<2, 1>(n)
                                    )
                                    - harmonic_sum<2, 3>(n)
                                    + (harmonic_sum<1>(n) + log_n(n, p))*(
                                        + pow(psi_1, 2)/2
                                        - 5./2.*harmonic_sum<4>(n)
                                        + 2* harmonic_sum<3, 1>(n)
                                    )
                                    + 4*harmonic_sum<3, 2>(n)
                                    + 7*harmonic_sum<4, 1>(n)
                                    + harmonic_sum<2, 2, 1>(n)
                                    - 6*harmonic_sum<3, 1, 1>(n)
                                )
                            )
                        )
                    )
                )
            )
        );
      }

      /* correction to f3 from three insertions of coulomb potential */
      double delta_f3_triple_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s/pi,3)*(
	    + pow(a1_eff(p),3)/64
	    + 4*beta(0, p.nl)*(
		+ pow(a1_eff(p),2)*(
		    - 11./64.
		    - n*psi_1/16
		    + harmonic_sum<1>(n)/32
		    + 3./32.*log_n(n, p)
		)
		+ 4*beta(0, p.nl)*(
		    + a1_eff(p)*(
			+ 3./8.
			- 3./16.*psi_1
			+ harmonic_sum<1>(n)*(
			    + 3./(16*n)
			    - 5./16.
			    - harmonic_sum<1>(n)/16
			    + log_n(n, p)/8
			)
			+ (
			    - 11./16.
			    + 3./16.*log_n(n, p)
			)*log_n(n, p)
			+ n*(
			    + 3./16.*harmonic_sum<3>(n)
			    - harmonic_sum<2,1>(n)/8
			    + n*(
				- 5./16.*harmonic_sum<4>(n)
				+ harmonic_sum<3,1>(n)/4
			    )
			    + psi_1*(
				+ 3./8.
				+ n*psi_1/16
				- log_n(n, p)/4
			    )
			    + zeta_3/16
			)
		    )
		    + 4*beta(0, p.nl)*(
			- 1./8.
			- 3./4.*zeta_3
			+ psi_1/2
			+ 5./4.*harmonic_sum<3>(n)
			- harmonic_sum<1,2>(n)
			- 5./8.*harmonic_sum<2,1>(n)
			+ (
			    + 11./8.*harmonic_sum<2>(n)
			    - 2*harmonic_sum<1,1>(n)
			    + harmonic_sum<1>(n)*(
				- 1./2.
				+ 11./8.*harmonic_sum<1>(n)
				+ 3./8.*log_n(n, p)
			    )
			)/n
			+ log_n(n, p)*(
			    + 3./4.
			    - 3./8.*psi_1
			    + (
				- 11./16.
				+ log_n(n, p)/8
			    )*log_n(n, p)
			)
			+ harmonic_sum<1>(n)*(
			    + 1./2.
			    + 6./6.*zeta_2
			    - 13./8.*psi_1
			    + harmonic_sum<1>(n)*(
				+ 1./16.
				- harmonic_sum<1>(n)/8
				- log_n(n, p)/8
			    )
			    + (
				- 5./8.
				+ log_n(n, p)/8
			    )*log_n(n, p)
			)
			+ n*(
			    - pow(pi,4)/120
			    - 3./16.*harmonic_sum<4>(n)
			    + 3./4.*harmonic_sum<3,1>(n)
			    + harmonic_sum<2,1>(n)/4*(
				+ 1
				- harmonic_sum<1>(n)
				- log_n(n, p)
			    )
			    + harmonic_sum<3>(n)*(
				- 1./4.
				+ 5./8.*harmonic_sum<1>(n)
				+ 3./8.*log_n(n, p)
			    )
			    + psi_1*(
				- 1./4.
				+ 7./16.*psi_1
				+ (
				    + 1
				    + harmonic_sum<1>(n)
				)*harmonic_sum<1>(n)/4
				+ (
				    + 3./4.
				    - log_n(n, p)/4
				)*log_n(n, p)
			    )
			    - zeta_3/4
			    - (harmonic_sum<1>(n)*zeta_3)/8
			    + (log_n(n, p)*zeta_3)/8
			    + n*(
				+ zeta_2/4*harmonic_sum<3>(n)
				- 15./8.*harmonic_sum<5>(n)
				+ harmonic_sum<3,2>(n)
				+ 2*harmonic_sum<4,1>(n)
				- harmonic_sum<3,1,1>(n)
				+ (5./8.*harmonic_sum<4>(n) - harmonic_sum<3,1>(n)/2)*(
				    + 1
				    + harmonic_sum<1>(n)
				    - log_n(n, p)
				)
				+ psi_1*(
				    - 9./8.*harmonic_sum<3>(n)
				    + harmonic_sum<2,1>(n)/2
				    + psi_1/8*(
					- 1
					- harmonic_sum<1>(n)
					+ log_n(n, p)
				    )
				    + zeta_3/8
				)
				- zeta_2*zeta_3/4
				- zeta_5/8
				+ n*(
				    + pow(pi,6)/1680
				    - zeta_3*zeta_3/4
				    + 2*zeta_2*harmonic_sum<4>(n)
				    + 11./4.*harmonic_sum<6>(n)
				    - zeta_2*harmonic_sum<3,1>(n)
				    - 2*harmonic_sum<3,3>(n)
				    - 7./2.*harmonic_sum<4,2>(n)
				    - 5*harmonic_sum<5,1>(n)
				    + harmonic_sum<3,1,2>(n)
				    + harmonic_sum<3,2,1>(n)
				    + 3*harmonic_sum<4,1,1>(n)
				    + harmonic_sum<3>(n)*(
					+ harmonic_sum<3>(n)/4
					+ zeta_3/2
				    )
				)
			    )
			)
		    )
		)
	    )
	);
      }

      /* correction to f3 from double insertion of coulomb and delta potential */
      double delta_f3_Coulomb_delta(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::a_ep;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s,3)/pi*pow(c_F,2)*(
	    - a_ep<1>(p.nl)/18
	    + a1_eff(p)*(
		- 175./432.
		- 8./(3*n)
		+ 4./3.*harmonic_sum<1>(n)
		- 4./3.*log_n(n, p)
	    )
	    + 4*beta(0,p.nl)*(
		+ 43./81.
		+ 8./(3*n)
		+ 67./36.*zeta_2
		+ 7./3.*psi_1
		+ log_n(n, p)*(
		    + 1./8.
		    - 16./(3*n)
		    - 7./3.*log_n(n, p)
		    + 4./3.*n*psi_1
		)
		+ harmonic_sum<1>(n)*(
		    - 29./72.
		    - 2./(3*n)
		    + harmonic_sum<1>(n)/3
		    + 2*log_n(n, p)
		    - 4./3.*n*psi_1
		)
		+ n*(
		    + 29*psi_1/36
		    - 8./3.*(harmonic_sum<3>(n)- zeta_3)
		    + 4./3.*(harmonic_sum<2,1>(n)- 2*zeta_3)
		)
	    )
	);
      }

      /* correction to f3 from double insertion of coulomb and 1/r^2 potential */
      double delta_f3_Coulomb_r2inv(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::b;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s,3)/pi*c_F*(
	    + a1_eff(p)*(
		+ c_F*(
		    - 1./4.
		    - 3./(4*n)
		    + 3./8.*harmonic_sum<1>(n)
		    - 3./8.*log_n(n, p)
		)
		+ c_A*(
		    + 11./16.
		    + 3./(2*n)
		    - 3./4.*harmonic_sum<1>(n)
		    + 3./4.*log_n(n, p)
		)
	    )
	    + 4*beta(0,p.nl)*(
		+ c_A*(
		    - 5./8.
		    - 7./(4*n)
		    - zeta_2
		    - 2*psi_1
		    + harmonic_sum<1>(n)*(
			+ 5./8.
			+ 3./(2*n)
			- harmonic_sum<1>(n)/2
			- log_n(n, p)
			+ n*psi_1
		    )
		    + log_n(n, p)*(
			+ 7./8.
			+ 3./n
			+ 3./2.*log_n(n, p)
			- n*psi_1
		    )
		    + n*(
			- 5./4.*psi_1
			+ 2*(harmonic_sum<3>(n)- zeta_3)
			- (harmonic_sum<2,1>(n)- 2*zeta_3)
		    )
		)
		+ c_F*(
		    + 1./4.
		    + 7./(8*n)
		    + zeta_2/2
		    + psi_1
		    + log_n(n, p)*(
			- 1./4.
			- 3./(2*n)
			- 3./4.*log_n(n, p)
			+ n*psi_1/2
		    )
		    + harmonic_sum<1>(n)*(
			- 1./4.
			- 3./(4*n)
			+ harmonic_sum<1>(n)/4
			+ log_n(n, p)/2
			- n*psi_1/2
		    )
		    + n*(
			+ psi_1/2
			- (harmonic_sum<3>(n)- zeta_3)
			+ (harmonic_sum<2,1>(n)- 2*zeta_3)/2
		    )
		)
	    )
	);
      }

      /* correction to f3 from double insertion
	 of coulomb and momentum-dependent p-potential */
      double delta_f3_Coulomb_p2(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::a_ep;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s,3)/pi*pow(c_F,2)*(
	    + a_ep<1>(p.nl)/12
	    + a1_eff(p)*(
		+ 29./12.
		- 5./(4*pow(n,2))
		+ 4./n
		- 2*harmonic_sum<1>(n)
		+ 2*log_n(n, p)
	    )
	    + 4*beta(0,p.nl)*(
		- 2./3.
		+ 9./(8*pow(n,2))
		- 2./n
		- 67./24.*zeta_2
		- (7./2. - 1./n + 3*n)*psi_1
		+ harmonic_sum<1>(n)*(
		    + 1./2.
		    - 3./(2*pow(n,2))
		    + 1./n
		    - harmonic_sum<1>(n)/2
		    - 3*log_n(n, p)
		    + 2*n*psi_1
		)
		+ log_n(n, p)*(
		    7./2.
		    - 5./(2*pow(n,2))
		    + 8./n
		    + 7./2.*log_n(n, p)
		    - 2*n*psi_1
		)
		+ n*(
		    + 4*(harmonic_sum<3>(n)- zeta_3)
		    - 2*(harmonic_sum<2,1>(n)- 2*zeta_3)
		)
	    )
	);
      }

      /* correction to f3 from double insertion
	 of coulomb potential and kinetic correction */
      double delta_f3_Coulomb_kinetic(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::a_ep;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s,3)/pi*pow(c_F,2)*(
	    + a_ep<1>(p.nl)/24
	    + a1_eff(p)*(
		+ 37./48.
		- 25./(32*pow(n,2))
		+ 5./(4*n)
		- 5./8.*harmonic_sum<1>(n)
		+ 5./8.*log_n(n, p)
	    )
	    + 4*beta(0,p.nl)*(
		+ 1./24.
		+ 5./(8*pow(n,2))
		- 1./(8*n)
		- 43./48.*zeta_2
		- (
		    + 3./4.
		    - 5./(8*n)
		    + 3./4.*n
		)*psi_1
		+ harmonic_sum<1>(n)*(
		    - 1./8.
		    - 15./(16*pow(n,2))
		    - 1./(4*n)
		    + n*psi_1/2
		)
		+ log_n(n, p)*(
		    + 9./8.
		    - 25./(16*pow(n,2))
		    + 5./(2*n)
		    - harmonic_sum<1>(n)
		    + log_n(n, p)
		    - n*psi_1/2
		)
		+ n*(
		    + (harmonic_sum<3>(n)- zeta_3)
		    - (harmonic_sum<2,1>(n)- 2*zeta_3)/2
		)
	    )
	);
      }

      /* correction to f3 from insertion of delta potential */
      double delta_f3_delta(int n, parameters const & p){
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	double log_m = log(p.mu/p.m);
	using Wilson_coefficients::vm1_ep;
	using Wilson_coefficients::vq1_ep;
	return pow(p.alpha_s,3)/pi*(
	    + pow(c_F,2)*(
		+ vm1_ep(p.nl)/8
		+ vq1_ep(p.nl)/12
		+ T_f*(
		    + 1./15.
		    + 4./(15*n)
		    - 2./15.*harmonic_sum<1>(n)
		    + p.nl*(
			+ 10./81.
			+ 40./(27*n)
			- 20./27.*harmonic_sum<1>(n)
			+ 20./27.*log_n(n, p)
		    )
		    + 2./15.*log_n(n, p)
		    + 1./15.*log_m
		)
		+ 4*beta(0,p.nl)*(
		    + 355./324.
		    - 5./(6*pow(n,2))
		    + 1./n
		    + 19./36.*zeta_2
		    + harmonic_sum<1>(n)*(
			- 1./3.
			+ 4./(3*n)
			- harmonic_sum<1>(n)/3
			+ 2./3.*log_n(n, p)
		    )
		    + (
			1./3.
			- 4./(3*n)
			- log_n(n, p)/3
		    )*log_n(n, p)
		    - psi_1/3
		)
	    )
	    + pow(c_F,3)*(
		- 137./36.
		+ 35./(12*pow(n,2))
		- 133./72.*zeta_2
		+ harmonic_sum<1>(n)*(
		    + 3./2.
		    - 14./(3*n)
		    + 7./6.*harmonic_sum<1>(n)
		    - 7./3.*log_n(n, p)
		    - 1./3.*log_m
		)
		+ log_n(n, p)*(
		    - 3./2.
		    + 14./(3*n)
		    + 7./6.*log_n(n, p)
		    + 1./3.*log_m
		)
		+ (
		    - 25./6.
		    + 2./3.*log_m
		)/n
		+ 1./12.*log_m * (1 + log_m)
		+ 7./6.*psi_1
	    )
	    + c_A*pow(c_F,2)*(
		+ 1337./648.
		- 55./(24*pow(n,2))
		+ 209./144.*zeta_2
		+ harmonic_sum<1>(n)*(
		    + 97./108.
		    + 11./(3*n)
		    - 11./12.*harmonic_sum<1>(n)
		    + 11./6.*log_n(n, p)
		    - 1./2.*log_m
		)
		+ log_n(n, p)*(
		    - 97./108.
		    - 11./(3*n)
		    - 11./12.*log_n(n, p)
		    + 1./2.*log_m
		)
		+ (
		    - 95./108.
		    + log_m
		)/n
		- 5./9.*log_m
		+ 1./8.*pow(log_m,2)
		- 11./12.*psi_1
	    )
	);
      }

      /* correction to f3 from insertion 1/r^2 potential */
      double delta_f3_r2inv(int n, parameters const & p){
	using Wilson_coefficients::b;
	using Wilson_coefficients::b_ep;
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	return pow(p.alpha_s,3)/pi*c_F/(144*n)*(
	    - 24*n*b_ep<2>(p.nl)
	    + 48*b<2>(p.nl)*(
		- 6
		- 4*n
		+ 3*n*harmonic_sum<1>(n)
		- 3*n*log_n(n, p)
		+ n*log(2)
	    )
	    - 8*c_A*(c_A + 2*c_F)*(
		- 72
		- 32*n
		+ 54*n*zeta_2
		- 24*n*pow(harmonic_sum<1>(n),2)
		- 48*(2 + n)*log_n(n, p)
		- 24*n*pow(log_n(n, p),2)
		+ 48*harmonic_sum<1>(n)*(2+ n + n*log_n(n, p))
		- 16*n*log(2)
		+ 8*n*pow(log(2),2)
		- 24*n*psi_1
	    )
	    + 4*beta(0,p.nl)*(
		+ 12*b_ep<1>(p.nl)*n
		+ b<1>(p.nl)*(
		    - 216
		    - 120*n
		    + 186*n*zeta_2
		    - 72*n*pow(harmonic_sum<1>(n),2)
		    - 144*(2 + n)*log_n(n, p)
		    - 72*n*pow(log_n(n, p),2)
		    + 144*harmonic_sum<1>(n)*(2 + n + n*log_n(n, p))
		    - 24*n*log(2)
		    - 72*n*psi_1
		)
	    )
	);
      }

      /* correction to f3 from insertion of momentum-dependent p-potential */
      double delta_f3_p2(int n, parameters const & p){
	double const psi_1 = zeta_2 - harmonic_sum<2>(n);
	using Wilson_coefficients::vp1_ep;
	return pow(c_F/n,2)*pow(p.alpha_s,3)/pi*(
	    + 25./36.*T_f*p.nl
	    + c_A*(
		- 107./144.
		- harmonic_sum<1>(n)
		- 5./3.*log_n(n, p)
	    )
	    + 4*beta(0, p.nl)*(
		+ 1./8.
		- 3./8.*harmonic_sum<1>(n)
		- 5./8.*log_n(n, p)
	    )
	    + n*(
		- 20./9.*T_f*p.nl
		+ 4*beta(0, p.nl)*(
		    + 3./2.
		    + psi_1/4
		    - 2*(harmonic_sum<1>(n) - log_n(n, p))
		)
		+ c_A*(
		    + 67./9.
		    + 2./3.*psi_1
		    - 16./3.*(harmonic_sum<1>(n) - log_n(n, p))
		)
		+ n*(
		    + vp1_ep(p.nl)/12
		    + p.nl*T_f*(
			- 40./27.
			+ 10./9.*(harmonic_sum<1>(n) - log_n(n, p))
		    )
		    + 4*beta(0, p.nl)*(
			+ 5./6.
			- 19./24.*zeta_2
			+ psi_1/2
			+ harmonic_sum<1>(n)*(
			    - 1
			    + harmonic_sum<1>(n)/2
			    - log_n(n, p)
			)
			+ (1 + log_n(n, p)/2)*log_n(n, p)
		    )
		    + c_A*(
			+ 122./27.
			- 19./9.*zeta_2
			+ 4./3.*psi_1
			+ harmonic_sum<1>(n)*(
			    - 79./18.
			    + 4./3.*harmonic_sum<1>(n)
			    - 8./3.*log_n(n, p)
			)
			+ log_n(n, p)*(79./18. + 4./3.*log_n(n, p))
		    )
		)
	    )
	);
      }

      double delta_f3_Higgs(int n, parameters const & p){
	return c_F/2*p.alpha_s/(4*pi)*p.Yukawa_t2*pow(p.m/p.m_Higgs,2)*(
	    2*log_n(n, p) + 1 + 4./n - 2*harmonic_sum<1>(n)
	);
      }

      double delta_f3_us_const(int n);

      /* ultra-soft correction to f3 */
      double delta_f3_ultrasoft(int n, parameters const & p){
	double log_m = log(p.mu/p.m);
	double P = harmonic_sum<1>(n) - 2./n + log(c_F/n);
	return
	  pow(p.alpha_s,3)/pi*(
	      + pow(c_A,3)*(
		  - 3./4.*log_m
		  + log(p.alpha_s)
	      )
	      + c_F*(
		  + pow(c_A,2)*(
		      + (
			  + 52./9. - 4*P - 8./3.*log(2) - 2*log(p.alpha_s)
		      )*log(p.alpha_s)
		      + log_m*(
			  - 11./3. - 5./6.*log_m + 8./3.*P
			  + 5./3.*log(2) + 8./3.*log(p.alpha_s)
		      )
		  )
		  + c_F*(
		      + (
			  + c_A*(
			      6 - 10./(3*pow(n,2)) - 32./3.*P - 4./3.*log(2)
			      - 16./3.*log(p.alpha_s)
			  )
			  + c_F*(
			      - 52./9. - 4./(3*pow(n,2)) - 16./3.*P + 8*log(2)
			      - 8./3.*log(p.alpha_s)
			  )
		      )*log(p.alpha_s)
		      + log_m*(
			  + c_F*(
			      + 5 - log_m/3 + 2./(3*pow(n,2)) + 8./3.*P - 6*log(2)
			      + 8./3.*log(p.alpha_s)
			  )
			  + c_A*(
			      -3./2. - 11./6.*log_m + 5./(3*pow(n,2)) +
			      20./3.*P + log(2)/3 + 20./3.*log(p.alpha_s)
			  )
		      )
		  )
	      )
	      + delta_f3_us_const(n)
	  );
      }

      double delta_f3_us_const(int n){
	if((n < 1) || (n > 6)){
	  throw std::out_of_range(
	      "ultrasoft correction fo wave function unknown for n = "
	      + std::to_string(n)
	      + " (not in range (1...6))"
	  );
	}
	static constexpr auto log_tab = std::array<double,6>{{
	    353.06,
	    256.62,
	    224.26,
	    206.88,
	    195.48,
	    187.16
	  }};
	return log_tab[n-1];
      }
    }
  }
}
