/**
 * @file
 * @brief Implementation of conversion to the pole mass scheme
 *
 * @see scheme_conversion.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.0
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "scheme_conversion.hpp"
#include "mass_schemes.hpp"
#include "energy_levels.hpp"
#include "array.hpp"
#include "hypergeometric.hpp"
#include "dilog.hpp"

#include "QCD/colour.hpp"
#include "QCD/beta.hpp"
#include "zeta.hpp"
#include "Wilson_coefficients.hpp"

namespace QQbar_threshold{
  using QCD::T_f;
  using QCD::c_F;
  using QCD::c_A;
  using QCD::beta;
  using Wilson_coefficients::a;
  using namespace zeta;

  double bottom_pole_mass(
      double mu, double m, order o,
      options opt
  ){
    opt.mass_scheme.name = to_shifted_mass(opt.mass_scheme.name);
    return gen_bottom_parameters(mu, m, o, opt).m;
  }

  double top_pole_mass(
      double mu, double m, order o,
      options opt
  ){
    opt.mass_scheme.name = to_shifted_mass(opt.mass_scheme.name);
    return gen_top_parameters(mu, m, o, opt).m;
  }

  namespace detail{
    namespace{
      double delta_m_sum(parameters const & p){
        double result = 0;
        for(int i = 0; i <= p.order; ++i){
          result += detail::delta_m(i, p);
        }
        return result;
      }
    }

    double mass_shift(parameters const & p){
      return is_shifted_mass(p.mass_scheme)?delta_m_sum(p):0.;
    }

    double bottom_mass_shift(
        double mu, double m, order o,
        options opt
    ){
      return gen_bottom_parameters(mu, m, o, std::move(opt)).m - m;
    }

    double top_mass_shift(
        double mu, double m, order o,
        options opt
    ){
      return gen_top_parameters(mu, m, o, std::move(opt)).m - m;
    }

    namespace{
      template<scheme_name S, int order>
        double delta_m_light(parameters const & p);

    }

    /*****************************************************************************

     Mass shifts for potential subtracted mass:
     M. Beneke, Phys.Lett.B434:115-125,1998, hep-ph/98042417
     Charm corrections are given in arXiv:1411.3132 by M. Beneke, A. Maier,
     J. Piclum, T. Rauh

    *****************************************************************************/

    template<>
    double delta_m<PS, 0>(parameters const & p){
      return c_F*p.alpha_s*p.mass_scheme.scale/pi;
    }

    namespace{
      /********** Massive light quark corrections to PS/pole relation ***********
       *                                                                        *
       * Equation numbers refer to                                              *
       * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132   *
       * \cite Beneke:2014pta						      *
       *									      *
       **************************************************************************/

      /************************** basis integrals *******************************/

      // eq. (B.6)
      double int_log(double m, double mu_f){
        return 2*(log(m/mu_f) + 1);
      }

      // eq. (B.7)
      double int_log2(double m1, double m2, double mu_f){
        return 4*(
            2 + log(m1/mu_f) + log(m2/mu_f) + log(m1/mu_f)*log(m2/mu_f)
        );
      }

      // eq. (B.9)
      double int_Pi(double z){
        using hypergeometric::PFQ;
        return z/5*(
            + PFQ<2>({1,1}, {7./2.}, -z)
            + 1./3.*PFQ<3>({1,1,3./2.},{5./2.,7./2.},-z)
        );
      }

      // eq. (B.13)
      double d_int_Pi(double z){
        using hypergeometric::PFQ;
        return (
            3*(5 + (-2 + z)*PFQ<2>({1,1}, {7./2.}, -z))/(1 + z)
            - PFQ<3>({1,1,3./2.}, {5./2.,7./2.}, -z)
        )/30;
      }

      // eq. (B.10)
      double int_Pi_log(double z, double m, double mu_f){
        using hypergeometric::PFQ;
        return
          + 2*int_Pi(z)*log(m/mu_f)
          + 2./15.*z*(
              + PFQ<3>({1,1,3./2.},{5./2.,7./2.},-z)
              + 1./3.*PFQ<4>({1,1,3./2.,3./2.},{5./2.,5./2.,7./2.},-z)
          );
      }

      // eq. (B.8)
      double int_Xi(double zt){
        return 2*atan(sqrt(zt))/sqrt(zt) + log(1 + zt) - 2;
      }

      // eq. (B.11)
      double int_Pi2(double z){
        using dilog::Li2;
        return (
            z*(-9 + z*(48 + 785*z))
            + asinh(sqrt(z))*(
                - 6*sqrt(z*(1 + z))*(-3 + z*(17 + 110*z))
                + 9*(-1 + 5*z + 20*pow(z,3))*asinh(sqrt(z))
            )
            + pow(z,5./2.)*(
                - 108*zeta_2
                + 36*asinh(sqrt(z))*(log(z) - 2*log(-1 + sqrt(1 + z)))
                - 36*Li2(pow(sqrt(z) + sqrt(1 + z),-2))
                + 144*Li2(pow(sqrt(z) + sqrt(1 + z),-1))
            )
        )/(45*pow(z,3));
      }

      // eq. (B.3), up to a factor of alpha_s/pi*C_F*mu_f*alpha_s/(4*pi)
      template<>
        double delta_m_light<PS, 1>(parameters const & p){
        if(p.ml == 0) return 0;
        const double mu_f = p.mass_scheme.scale;
        const double z = pow(mu_f/(2*p.ml),2);
        return 4./3.*QCD::T_f*(int_Pi(z) + int_log(p.ml, mu_f) + 5./3.);
      }

      // eq. (B.4), up to a factor of alpha_s/pi*C_F*mu_f*(alpha_s/(4*pi))^2
      template<>
        double delta_m_light<PS, 2>(parameters const & p){
        if(p.ml == 0) return 0;
        using QCD::T_f;
        const double mu_f = p.mass_scheme.scale;
        const double mu_l = p.mu; /* subject to debate */
        const double z = pow(mu_f/(2*p.ml),2);
        /* Parameters for the charm effects in the NNLO Coulomb potential */
        /* Source: A. Hoang, hep-ph/0008102 */
        static constexpr double c2 = 0.470;
        static constexpr double d2 = 1.120;
        static constexpr double c1 = -0.8316040946513316;
        static constexpr double d1 = 1.8316040946513317;

        return 2*a<1>(p.nl)*delta_m_light<PS,1>(p)
          + 8./3.*T_f*4*beta(0, p.nl)*(
              + int_Pi_log(z, p.mu, mu_f)
              + int_log2(p.ml, p.mu, mu_f)
              + 5./3.*int_log(p.mu, mu_f)
          )
          + pow(4./3.*T_f,2)*(
              + int_Pi2(z) + 2*int_Pi_log(z, p.ml, mu_f) + 10./3.*int_Pi(z)
              + int_log2(p.ml, p.ml, mu_f) + 10./3.*int_log(p.ml, mu_f) + 25./9.
          )
          + 76./3.*T_f*(
              + c1*int_Xi(z/pow(c2,2)) + d1*int_Xi(z/pow(d2,2))
              + int_log(p.ml, mu_f) + 161./114. + 26./19.*zeta_3
          )
          + 16./3.*T_f*c_F*(2 + 3*log(mu_l/p.ml))*(1 - z*d_int_Pi(z));
      }
    }

    template<>
    double delta_m<PS, 1>(parameters const & p){
      const double log_mu = log(p.mu/p.mass_scheme.scale);
      return p.contributions.v_Coulomb[0]
        *c_F*p.alpha_s*p.mass_scheme.scale/pi*p.alpha_s/(4*pi)*(
            + a<1>(p.nl)
            + 2*4*beta(0, p.nl)*(1 + log_mu)
            + delta_m_light<PS, 1>(p)
        );
    }

    template<>
    double delta_m<PS, 2>(parameters const & p){
      const double log_mu = log(p.mu/p.mass_scheme.scale);
      return p.contributions.v_Coulomb[1]
        *c_F*p.alpha_s*p.mass_scheme.scale/pi*pow(p.alpha_s/(4*pi),2)*(
            + a<2>(p.nl)
            + 2*(
                + 2*a<1>(p.nl)*4*beta(0, p.nl)
                + 16*beta(1, p.nl)
            )*(1 + log_mu)
            + 4*pow(4*beta(0, p.nl),2)*(
                2 + 2*log_mu + pow(log_mu,2)
            )
            + delta_m_light<PS, 2>(p)
        );
    }

    /**
     *  delta_m<3,PS> from eq .37 of M. Beneke, Y. Kiyo, K. Schuller,
     *  Nucl.Phys.B714:67-90,2005, hep-ph/0501289 (note the typo in the
     *  last term 16 Pi^2 CA^3, where the factor l_ 1 is missing).
     */
    template<>
    double delta_m<PS, 3>(parameters const & p){
      const double log_mu = log(p.mu/p.mass_scheme.scale);
      return p.contributions.v_Coulomb[2]
        *c_F*p.alpha_s*p.mass_scheme.scale/pi*pow(p.alpha_s/(4*pi),3)*(
            + a<3>(p.nl)
            + 16*pow(pi,2)*pow(c_A,3)*(1 /*+ log(nu/p.mass_scheme.scale)*/) // nu = mass_scheme.scale
            + 8*pow(4*beta(0, p.nl),3)*(
                + 6
                + 6*log_mu
                + 3*pow(log_mu,2)
                + pow(log_mu,3)
            )
            + 10*4*beta(0, p.nl)*16*beta(1, p.nl)*(
                + 2
                + 2*log_mu
                + pow(log_mu,2)
            )
            + 2*64*beta(2, p.nl)*(1 + log_mu)
            + 6*a<2>(p.nl)*4*beta(0, p.nl)*(1 + log_mu)
            + 12*a<1>(p.nl)*pow(4*beta(0, p.nl),2)*(
                + 2
                + 2*log_mu
                + pow(log_mu,2)
            )
            + 4*a<1>(p.nl)*16*beta(1, p.nl)*(1 + log_mu)
        );
    }

    /**********************************************************************
     Upsilon(1S) mass scheme:
     A.H. Hoang, Z. Ligeti, A.V. Manohar, Phys.Rev.Lett.82:277-280,1999,
     hep-ph/9809423
    *********************************************************************/

    template<>
    double delta_m<OneS, 0>(parameters const & p){
      return - detail::E_LO(1, p)/2;
    }

    template<>
    double delta_m<OneS, 1>(parameters const & p){
      return - detail::E_LO(1, p)/2 * detail::e1_no_mass_shift(1, p);
    }

    template<>
    double delta_m<OneS, 2>(parameters const & p){
      using namespace detail;
      return - E_LO(1, p)/2 * (e2_no_mass_shift(1, p) - E_LO(1, p)/(2*p.m));
    }

    template<>
    double delta_m<OneS, 3>(parameters const & p){
      using namespace detail;
      return - E_LO(1, p)/2 * (
          + e3_no_mass_shift(1, p)
          - E_LO(1, p)/(2*p.m)*(
              2*e1_no_mass_shift(1, p) - p.alpha_s/pi*4*beta(0, p.nl)
          )
      );
    }

    /***********************************************************************
     MSbar scheme including light quark effects
     Bekavac et al. Nucl. Phys. B833 46-63, arXiv:0911.3356
    **********************************************************************/

    template<>
    double delta_m<MSbar, 0>(parameters const & p){
      double const log_m = log(p.mass_scheme.scale/p.m);
      return p.m*p.alpha_s/pi*(
	  + 4./3.
          + 2*log_m
      );
    }

    namespace{
      template<>
        double delta_m_light<MSbar,1>(parameters const & p){
        using dilog::Li2;
        if(p.ml == 0) return 0;
        double const x = p.ml/p.m;
        double const log_x = log(x);
        double const li_sum = Li2(x) + Li2(-x) + log_x*(log(1+x) + log(1-x));
        double const li_diff = 3*zeta_2 + Li2(-x) - Li2(x) - log_x*(log(1-x) - log(1+x));
        return c_F*T_f/2*(
            li_sum + x*(
                li_diff + x*(
                    - 3./2. - log_x + x*(
                        li_diff + (li_sum - pow(log_x,2) - zeta_2)*x
                    )
                )
            )
        );
      }

    }

    template<>
    double delta_m<MSbar, 1>(parameters const & p){
      double const log_m = log(p.mass_scheme.scale/p.m);
      double const log_mu = log(p.mass_scheme.scale/p.mu);
      return p.m*pow(p.alpha_s/pi,2)*(
          + 307./32. + 2*zeta_2 + 2./3.*log(2)*zeta_2 - zeta_3/6
          - 22./3.*log_mu
          + log_m*(
              + 509./36. + 47./6.*log_m - 11*log_mu
              + (-13./18. - log_m/3 + 2./3.*log_mu)*p.nl
          )
          + p.nl*(-71./144. + 4./9.*log_mu - zeta_2/3)
          + delta_m_light<MSbar, 1>(p)
      );
    }

    namespace{
      double z3(parameters const & p){
        double const x = p.ml/p.m;
        double const log_x = log(x);
        double const log_mu = log(p.mass_scheme.scale/p.m);
        return x*(
            -22.854111111111102 - 14.290444444444447*log_mu
            + p.nl*(
                + 0.9826666666666666
                + 1.0693333333333332*log_mu
                - 1.1013333333333333*log_x
            )
            + 18.049*log_x
            + x*(-4.6483333333333325
                 + log_mu*(5.370222222222221 - 0.43999999999999995*p.nl)
                 + 0.30033333333333334*p.nl
                 + x*(
                     -1.0221111111111112
                     + log_mu*(-1.5826666666666669 + 0.134*p.nl)
                     - 0.049333333333333326*x
                     - 0.03433333333333333*log_x
                 )
                 + 2.787555555555555*log_x
            )
        );
      }

      double del_z(parameters const & p){
        using dilog::Li2;
        double const x = p.ml/p.m;
        double const log_x = log(x);
        double const log_mu = log(p.mass_scheme.scale/p.m);
        /* for mu_light = mass_scheme.scale */
        double const log_mu_l = 2*log_mu - 2*log_x;
        double const li_sum = Li2(x) + Li2(-x) + log_x*(log(1+x) + log(1-x));
        double const li_diff =
          3*zeta_2 + Li2(-x) - Li2(x) - log_x*(log(1-x) - log(1+x));
        return (
            2*li_sum*(1 + 6*log_mu)
            + x*(
                + li_diff*(2 + 6*log_mu + 3*log_mu_l)
                + x*(
                    - 3 + 6*log_mu - 2*log_x - 6*log_mu_l*(2 + log_x)
                    + x*(
                        + li_diff*(2 - 6*log_mu + 9*log_mu_l)
                        + 2*(-1 + 6*log_mu - 6*log_mu_l)*x*(
                            pow(log_x,2) + zeta_2 - li_sum
                        )
                    )
                )
            )
        )/8;
      }

      template<>
        double delta_m_light<MSbar,2>(parameters const & p){
        if(p.ml == 0) return 0;
        return
          - z3(p) + c_F*c_F*T_f*del_z(p)
          + delta_m_light<MSbar,1>(p)*(
              + 4*beta(0, p.nl)*log(p.mu/p.mass_scheme.scale)
              + 2./3.*log(p.mu/p.m)
          );
      }
    }


    constexpr double a4 = 0.51747906167389938633; //a4 = Li_4(1./2.)

    template<>
    double delta_m<MSbar, 2>(parameters const & p){
      double const log_m = log(p.mass_scheme.scale/p.m);
      double const log_mu = log(p.mass_scheme.scale/p.mu);
      return p.m*pow(p.alpha_s/pi,3)*(
          + 8462917./93312. - 220./27.*a4 - 55./162.*pow(log(2),4)
          + 652841./6480.*zeta_2 - 575./27.*log(2)*zeta_2
          - 44./27.*pow(log(2),2)*zeta_2 + 58./27.*zeta_3
          - 1439./72.*zeta_2*zeta_3 - 3475./432.*zeta_4
          + 1975./216.*zeta_5
          + log_m*(
              + 93391./648. + 26*zeta_2 + 26./3.*log(2)*zeta_2 - 23./6.*zeta_3
              + log_m*(21715./216. + 1861./54.*log_m - 517./6.*log_mu)
              + log_mu*(-6517./36. + 121./2.*log_mu)
          )
          + log_mu*(
              - 3921./32. - 22*zeta_2 - 22./3.*log(2)*zeta_2 + 11./6.*zeta_3
              + 121./3.*log_mu
          )
          + p.nl*(
              - 231847./23328. + 8./27.*a4 + pow(log(2),4)/81 - 991./108.*zeta_2
              - 22./27.*log(2)*zeta_2 + 4./27.*pow(log(2),2)*zeta_2
              - 241./72.*zeta_3 + 305./108.*zeta_4
              + log_m*(
                  - 5171./324. - 17./3.*zeta_2
                  - 4./9.*log(2)*zeta_2 - 14./9.*zeta_3
                  + (1109./54. - 22./3.*log_mu)*log_mu
                  + log_m*(-385./36. - 86./27.*log_m + 80./9.*log_mu)
                  + p.nl*(
                      + 89./324. + 2./9.*zeta_2
                      + log_m*(13./54. + 2./27.*log_m - 2./9.*log_mu)
                      + (-13./27. + 2./9.*log_mu)*log_mu
                  )
              )
              + log_mu*(
                  + 1003./72. + 5*zeta_2 + 4./9.*log(2)*zeta_2 - zeta_3/9
                  - 44./9.*log_mu
                  + p.nl*(-71./216. + 4./27.*log_mu - 2./9.*zeta_2)
              ) + p.nl*(2353./23328. + 13./54.*zeta_2 + 7./54.*zeta_3)
          )
          + delta_m_light<MSbar, 2>(p)
      );
    }

    template<>
    double delta_m<MSbar, 3>(parameters const & p){
      double log_m = log(p.mass_scheme.scale/p.m);
      double const log_mu = log(p.mass_scheme.scale/p.mu);
      return p.m*pow(p.alpha_s/pi,4)*(
          + p.r4
          + log_mu*(
              - 112013363./62208. + 1210./9.*a4 + 605./108.*pow(log(2),4)
              - 7401571./4320.*zeta_2 + 6019./18.*log(2)*zeta_2
              + 242./9.*pow(log(2),2)*zeta_2 - 1123./36.*zeta_3
              + 15829./48.*zeta_2*zeta_3 + 38225./288.*zeta_4
              - 21725./144.*zeta_5
              + log_mu*(
                  + 141361./128. + 363./2.*zeta_2
                  + 121./2.*log(2)*zeta_2 - 121./8.*zeta_3
                  - 1331./6.*log_mu
              )
          )
          + log_m*(
              + 339048211./186624. - 4070./27.*a4 - 2035./324.*pow(log(2),4)
              + 24353117./12960.*zeta_2 - 3500./9.*log(2)*zeta_2
              - 814./27.*pow(log(2),2)*zeta_2 + 817531./13824.*zeta_3
              - 53243./144.*zeta_2*zeta_3 - 122995./864.*zeta_4
              + 50275./432.*zeta_5
              + log_m*(
                  + 152459./108.
                  + 1447./6.*zeta_2 + 1447./18.*log(2)*zeta_2 - 2737./72.*zeta_3
                  + log_m*(835903./1296. + 23159./144.*log_m - 20471./36.*log_mu)
                  + log_mu*(-267629./144. + 5687./8.*log_mu)
              )
              + log_mu*(
                  - 2443249./864.
                  - 429*zeta_2 - 143*log(2)*zeta_2
                  + 253./4.*zeta_3
                  + (78419./48. - 1331./4.*log_mu)*log_mu
              )
          )
          + p.nl*(
              + log_m*(
                  -275023./864. + 368./27.*a4 + 46./81.*pow(log(2),4)
                  - 1796711./6480.*zeta_2 + 130./27.*log(2)*zeta_2
                  + 118./27.*pow(log(2),2)*zeta_2
                  - 37805./432.*zeta_3 + 1439./72.*zeta_2*zeta_3
                  + 28655./432.*zeta_4 + 1475./216.*zeta_5
                  + log_m*(
                      - 1259743./5184.
                      - 2359./36.*zeta_2 - 76./9.*log(2)*zeta_2
                      - 539./36.*zeta_3
                      + (8378./27. - 1397./12.*log_mu)*log_mu
                      + log_m*(-8579./81. - 1685./72.*log_m + 4699./54.*log_mu)
                  )
                  + log_mu*(
                      + 1265221./2592.
                      + 239./2.*zeta_2 + 16*log(2)*zeta_2
                      + 131./6.*zeta_3
                      + log_mu*(-10291./36. + 121./2.*log_mu)
                  )
                  + p.nl*(
                      + 51229./3888. - 8./27.*a4 - pow(log(2),4)/81
                      + 43./3.*zeta_2 + 22./27.*log(2)*zeta_2
                      - 4./27.*pow(log(2),2)*zeta_2 + 1403./216.*zeta_3
                      - 175./54.*zeta_4
                      + log_mu*(
                          - 60019./2592. - 28./3.*zeta_2
                          - 4./9.*log(2)*zeta_2 - 14./9.*zeta_3
                          + log_mu*(
                              + 413./27.
                              + log_mu*(-11./3. + 2./27.*p.nl)
                              - 13./54.*p.nl
                          )
                          + p.nl*(89./324. + 2./9.*zeta_2)
                      )
                      + p.nl*(- 1301./11664. - 13./54.*zeta_2 - zeta_3/9)
                      + log_m*(
                          + 58567./5184.
                          + 44./9.*zeta_2 + 2./9.*log(2)*zeta_2 + 7./9.*zeta_3
                          + log_m*(428./81. + 41./36.*log_m - 119./27.*log_mu)
                          + log_mu*(-283./18. + 113./18.*log_mu)
                          + p.nl*(
                              - 89./648.
                              - zeta_2/9
                              + log_m*(-13./162. - log_m/54 + 2./27.*log_mu)
                              + (13./54. - log_mu/9)*log_mu
                          )
                      )
                  )
              )
              + log_mu*(
                  28859989./93312. - 352./27.*a4 - 44./81.*pow(log(2),4)
                  + 1730051./6480.*zeta_2 - 155./27.*log(2)*zeta_2
                  - 110./27.*pow(log(2),2)*zeta_2 + 24559./432.*zeta_3
                  - 1439./72.*zeta_2*zeta_3 - 23605./432.*zeta_4
                  + 1975./216.*zeta_5
                  + log_mu*(
                      - 111439./576. - 209./4.*zeta_2
                      - 22./3.*log(2)*zeta_2 + 11./6.*zeta_3
                      + 121./3.*log_mu
                  )
                  + p.nl*(
                      - 625889./46656. + 8./27.*a4 + pow(log(2),4)/81
                      - 767./54.*zeta_2 - 22./27.*log(2)*zeta_2
                      + 4./27.*pow(log(2),2)*zeta_2
                      - 395./72.*zeta_3 + 305./108.*zeta_4
                      + log_mu*(
                          + 8969./864.
                          + 13./3.*zeta_2 + 2./9.*log(2)*zeta_2 - zeta_3/18
                          - 22./9.*log_mu
                          + p.nl*(-71./432. + 4./81.*log_mu - zeta_2/9)
                      )
                      + p.nl*(2353./23328. + 13./54.*zeta_2 + 7./54.*zeta_3)
                  )
              )
          )
      );
    }

    //enforce template instantiation
    template double delta_m<PS,0>(parameters const & p);
    template double delta_m<PS,1>(parameters const & p);
    template double delta_m<PS,2>(parameters const & p);
    template double delta_m<PS,3>(parameters const & p);

    template double delta_m<OneS,0>(parameters const & p);
    template double delta_m<OneS,1>(parameters const & p);
    template double delta_m<OneS,2>(parameters const & p);
    template double delta_m<OneS,3>(parameters const & p);

    template<scheme_name S>
    double delta_m(int order, parameters const & p){
      switch(order){
      case 0: return delta_m<S,0>(p);
      case 1: return delta_m<S,1>(p);
      case 2: return delta_m<S,2>(p);
      case 3: return delta_m<S,3>(p);
        //add further cases if needed
      default:
        throw std::out_of_range(
            "order " + std::to_string(order)
            + " out of range (0,...,3) for mass shift"
        );
      }
    }

    double delta_m(int order, parameters const & p){
      switch(p.mass_scheme.name){
      case PSshift:
      case PS: return delta_m<PS>(order, p);
      case OneSshift:
      case OneS: return delta_m<OneS>(order, p);
      case MSbar: return delta_m<MSbar>(order, p);
      case Pole:
        return 0.0;
      default:
        throw not_implemented(
            "mass shift for scheme " + std::to_string(p.mass_scheme.name)
        );
      }
    }
  }
}
