#ifdef NDEBUG
#undef NDEBUG
#endif

#include <complex>
#include <cassert>
#include <array>
#include <iostream>

#include "residues.hpp"
#include "mass_schemes.hpp"
#include "scheme_conversion.hpp"

#define TSTB(name,i) \
  std::cout << #name << ": "; \
  diff = name - results_bottom[i]; \
  std::cout << diff; \
  assert(std::abs(diff)/std::abs(results_bottom[i]) < ep);	\
  std::cout << " (ok)\n";

#define TSTT(name,i) \
  std::cout << #name << ": "; \
  diff = name - results_top[i]; \
  std::cout << diff; \
  assert(std::abs(diff)/std::abs(results_top[i]) < ep);	\
  std::cout << " (ok)\n";

int main(){
  using namespace QQbar_threshold::pert_order;
  using QQbar_threshold::Pole;
  using QQbar_threshold::PS;
  using QQbar_threshold::PSshift;
  using QQbar_threshold::mu_f_top;
  using QQbar_threshold::mu_f_bottom;
  using QQbar_threshold::top_options;
  using QQbar_threshold::bottom_options;
  using QQbar_threshold::SM;
  auto opt = top_options();
  opt.alpha = 1/137.03599907430635;
  opt.m_Higgs = 125.9;
  opt.ml = 0;
  opt.beyond_QCD = SM::none;
  using QQbar_threshold::bbbar_residue;
  using QQbar_threshold::ttbar_residue;
  double diff;
  double constexpr ep = 1e-3;
  std::cout << "target precision: " << ep << "\n";

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, pole scheme:\n";

  opt.mass_scheme = {Pole, 0};
  opt.expand_s = true;

  auto results_top = std::array<double, 4>{{
      891.0300982027387,
      1072.687310692808,
      1325.6854245044858,
      1326.888127006273
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_residue(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, PS scheme:\n";

  opt.mass_scheme = {PS, mu_f_top};

  results_top = std::array<double, 4>{{
      891.0300982027387,
      1072.687310692808,
      1330.9062656360807,
      1332.9408694559568
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_residue(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, PSshift scheme:\n";

  opt.mass_scheme = {PSshift, mu_f_top};

  results_top = std::array<double, 4>{{
      906.7845727621443,
      1095.9645138985716,
      1359.0401288803498,
      1361.335171408792
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_residue(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, QED effects:\n";
  opt.beyond_QCD = SM::QED;

  results_top = std::array<double, 4>{{
      906.7845727621443,
      1151.1235635481266,
      1410.9423284981858,
      1427.0809011940546
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_residue(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "top, QED + Higgs effects:\n";
  opt.beyond_QCD |= SM::Higgs;

  results_top = std::array<double, 4>{{
      906.7845727621443,
      1151.1235635481266,
      1468.0952670501904,
      1528.4554528208314
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTT(ttbar_residue(1, 80, 175, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, Pole scheme:\n";
  opt = bottom_options();
  opt.beyond_QCD = SM::none;
  opt.expand_s = true;

  opt.mass_scheme = {Pole, 0};

  auto results_bottom = std::array<double, 4>{{
      0.014350596995008716,
      0.032837144730325916,
      0.0593499895997149,
      0.08484031921348857
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_residue(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PS scheme:\n";

  opt.mass_scheme = {PS, mu_f_bottom};

  results_bottom = std::array<double,4>{{
      0.014350596995008716,
      0.032837144730325916,
      0.05987023712263693,
      0.08592313512860067
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_residue(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PSshift scheme:\n";

  opt.mass_scheme = {PSshift, mu_f_bottom};

  results_bottom = std::array<double,4>{{
      0.015968604394871067,
      0.038221405513639055,
      0.07203647541518796,
      0.1065040363671398
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_residue(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PSshift scheme (mc = 0.9):\n";

  opt.mass_scheme = {PSshift, mu_f_bottom};
  opt.ml = 0.9;
  opt.expand_s = false;

  results_bottom = std::array<double,4>{{
      0.016049823776082113,
      0.04030511839307066,
      0.07975064522186878,
      0.11708518069288575
    }};

  for(order o = LO; o <= N3LO; ++o){
    TSTB(bbbar_residue(2, 5, 5, o, opt), o);
  }

  std::cout << "--------------------------------------------------\n";
  std::cout << "bottom, PSshift scheme (mc = 0.9) with massive double insertions:\n";

  QQbar_threshold::options opt2 = opt;
  opt2.double_light_insertion = true;

  results_bottom = std::array<double,4>{{
      0,
      0,
      0.00019753667814820242,
      0.00014603913518430878
    }};

  for(order o = LO; o <= N3LO; ++o){
    double diff;
    diff = std::abs(
	bbbar_residue(2, 5, 5, o, opt2)
	- bbbar_residue(2, 5, 5, o, opt)
	- results_bottom[o]
    );
    std::cout << o << ": " << diff;
    assert(diff < ep/10);
    std::cout << " (ok)\n";
  }

}
